from typing import Mapping
from typing import Optional
from typing import Type
from typing import Union

from ewokscore.task import Task

from ..gui.owwidgets.types import OWEwoksBaseWidget
from ..gui.owwidgets.types import is_ewoks_widget_class
from ..gui.owwidgets.types import is_native_widget_class
from ..gui.workflows.task_wrappers import execute_ewoks_owwidget
from ..gui.workflows.task_wrappers import execute_native_owwidget


def execute_task(
    task_cls: Union[Type[Task], Type[OWEwoksBaseWidget]],
    inputs: Optional[Mapping] = None,
    timeout: int = 60,
    **widget_init_params,
) -> dict:
    """Execute the task (use the orange widget or ewoks task class) and return the results"""
    if is_ewoks_widget_class(task_cls):
        return execute_ewoks_owwidget(
            task_cls, inputs=inputs, timeout=timeout, **widget_init_params
        )
    if is_native_widget_class(task_cls):
        return execute_native_owwidget(
            task_cls, inputs=inputs, timeout=timeout, **widget_init_params
        )
    if issubclass(task_cls, Task):
        task = task_cls(inputs=inputs)
        task.execute()
        return task.get_output_values()
    raise TypeError("task")



import warnings

from .widgets.data_viewer import DataViewer  # noqa F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.widgets.data_viewer' module.",
    DeprecationWarning,
    stacklevel=2,
)



from ._oasys_patch import oasys_patch
from .orange_utils.signal_manager import patch_signal_manager
from .owwidgets.summarizers import summarize  # noqa: F401
from .workflows.owscheme import patch_parse_ows_stream

oasys_patch()
patch_parse_ows_stream()
patch_signal_manager()



"""
Abstract base class for Ewoks-Orange widgets.
"""

from __future__ import annotations

import logging
import warnings
from typing import Any
from typing import Callable
from typing import List
from typing import Mapping
from typing import Optional

from AnyQt import QtWidgets
from ewokscore import missing_data
from ewokscore.variable import value_from_transfer

from ...orange_version import ORANGE_VERSION

# OWBaseWidget: lowest level Orange widget base class
# OWWidget: highest level Orangewidget base class.
if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets.widget import OWWidget

    OWBaseWidget = OWWidget
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets.widget import OWWidget
    from orangewidget.widget import OWBaseWidget
else:
    from orangewidget.widget import OWBaseWidget

    OWWidget = OWBaseWidget

from ..orange_utils._signals import get_signal
from ..orange_utils.orange_imports import OWBaseWidget
from ..orange_utils.orange_imports import OWWidget
from ..orange_utils.signals import Output
from ..utils import invalid_data
from ..utils.events import scheme_ewoks_events
from ..utils.model import _get_model_default_values
from .meta import OWEwoksWidgetMetaClass
from .meta import ow_build_opts

_logger = logging.getLogger(__name__)


class OWEwoksBaseWidget(OWWidget, metaclass=OWEwoksWidgetMetaClass, **ow_build_opts):
    """
    Abstract base class connecting Ewoks tasks with Orange workflow widgets.

    This class manages inputs (default and dynamic), constructs task arguments,
    and provides hooks for executing tasks and propagating outputs.

    Default input values are saved in the workflow file.
    Typically default input values are provided by the user through a widget component.

    Dynamic input values are not saved in the workflow file.
    Typically dynamic input values are send from the output if upstream tasks and wrapped
    by a `Variable` to handle things like Ewoks tasks output caching.

    Subclasses must implement:

    - methods: get_task_outputs, _execute_ewoks_task
    - properties: task_succeeded, task_done, task_exception
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize base widget internals.

        :param args: Positional args forwarded to parent.
        :param kwargs: Keyword args forwarded to parent.
        """
        super().__init__(*args, **kwargs)
        self.__dynamic_inputs = dict()
        self.__task_output_changed_callbacks: List[Callable[[], None]] = [
            self.task_output_changed
        ]
        self.__post_task_exception: Optional[Exception] = None

    # --- Control and Main area --------------------------------------------------------------

    def _init_control_area(self) -> None:
        """
        Initialize control area typically used for input controls and action buttons.

        Adds "Trigger" and "Execute" buttons wired to execution entry points.
        """
        layout = self._get_control_layout()

        trigger = QtWidgets.QPushButton("Trigger")
        execute = QtWidgets.QPushButton("Execute")

        layout.addWidget(trigger)
        trigger.released.connect(self.execute_ewoks_task)
        self._trigger_button = trigger

        layout.addWidget(execute)
        execute.released.connect(self.execute_ewoks_task_without_propagation)
        self._execute_button = execute

    def _init_main_area(self):
        """
        Initialize main area typically used to display results.
        """
        self._get_main_layout()

    def _get_control_layout(self):
        """
        Get or create the control area layout.

        :return: Qt layout instance for control area.
        """
        layout = self.controlArea.layout()
        # sp = self.controlArea.sizePolicy()
        # sp.setVerticalPolicy(QtWidgets.QSizePolicy.Expanding)
        # self.controlArea.setSizePolicy(sp)
        # print("changed the size policy")
        if layout is None:
            layout = QtWidgets.QVBoxLayout()
            self.controlArea.setLayout(layout)
        return layout

    def _get_main_layout(self):
        """
        Get or create the main area layout.

        :raises RuntimeError: If the widget doesn't declare `want_main_area`.
        :return: Qt layout instance for main area.
        """
        if not self.want_main_area:
            raise RuntimeError(
                f"{type(self).__name__} must have class attribute `want_main_area = True`"
            )
        layout = self.mainArea.layout()
        if layout is None:
            layout = QtWidgets.QVBoxLayout()
            self.mainArea.setLayout(layout)
        return layout

    # --- Ewoks task inputs --------------------------------------------------------------

    @classmethod
    def get_input_names(cls):
        """
        Return Ewoks task input names for the bound task class.

        :return: Iterable of input name strings.
        """
        return cls.ewokstaskclass.input_names()

    def get_task_inputs(self) -> dict:
        """
        Merge default and dynamic inputs producing the inputs mapping used by tasks.

        :return: Mapping of input name -> Variable or value (may include missing markers).
        """
        inputs = self.get_default_input_values()
        inputs.update(self.__dynamic_inputs)
        return inputs

    def get_task_input_values(self) -> dict:
        """
        Return all task input values (dynamic or default when missing).

        :return: Dict of input name -> plain value.
        """
        return {k: self._extract_value(v) for k, v in self.get_task_inputs().items()}

    def get_task_input_value(
        self, name: str, default: Any = missing_data.MISSING_DATA
    ) -> Any:
        """
        Retrieve a single task input value by name, returning default if missing.

        :param name: Input name.
        :param default: Fallback when missing.
        :return: The extracted input value or default.
        """
        adict = self.get_task_inputs()
        try:
            value = adict[name]
        except KeyError:
            return default
        value = self._extract_value(value)
        if missing_data.is_missing_data(value):
            return default
        return value

    # --- Ewoks task default inputs (SAVED IN FILE) --------------------------------------------------------------

    def get_default_input_names(self, include_missing: bool = False) -> set:
        """
        Return input names that have default values (or all input names).

        :param include_missing: If True return all defined input names.
        :return: Set of input names.
        """
        self._deprecated_default_inputs()
        if include_missing:
            return set(self.get_input_names())
        else:
            return set(self._ewoks_default_inputs)

    def get_default_input_values(
        self, include_missing: bool = False, defaults: Optional[Mapping] = None
    ) -> dict:
        """
        Return default input values or a mapping including missing markers.

        :param include_missing: If True include all input names set to INVALIDATION_DATA initially.
        :param defaults: Optional mapping of default overrides.
        :return: Dict of input name -> value or missing marker.
        """
        self._deprecated_default_inputs()
        if include_missing:
            values = {
                name: invalid_data.INVALIDATION_DATA for name in self.get_input_names()
            }
        else:
            values = dict()

        input_model = self.ewokstaskclass.input_model()

        if input_model is not None:
            # remove Values set to None. This defines "invalid downstream" in Orange.
            explicit_values = dict(
                filter(
                    lambda pair: pair[1] is not None,
                    _get_model_default_values(input_model).items(),
                )
            )
        else:
            explicit_values = {}
        values.update(explicit_values)

        if defaults:
            values.update(defaults)
        values.update(self._ewoks_default_inputs)

        return {name: invalid_data.as_missing(value) for name, value in values.items()}

    def get_default_input_value(self, name: str, default: Any = None) -> Any:
        """
        Get a default input value saved in the widget settings.

        :param name: Input name.
        :param default: Fallback if the value is not present.
        :return: The default value or provided fallback.
        """
        return self._ewoks_default_inputs.get(name, default)

    def set_default_input(self, name: str, value: Any) -> None:
        """
        Set or remove a default input.

        :param name: Input name.
        :param value: Input value. If it's invalidation data the default is removed.
        """
        if invalid_data.is_invalid_data(value):
            _logger.debug("ewoks widget: remove default input %r", name)
            _ = self._ewoks_default_inputs.pop(name, None)
        else:
            _logger.debug("ewoks widget: set default input %r = %s", name, value)
            self._ewoks_default_inputs[name] = value

    def update_default_inputs(self, **inputs) -> None:
        """
        Batch-set default inputs.

        :param inputs: name=value pairs to set as defaults.
        """
        for name, value in inputs.items():
            self.set_default_input(name, value)

    def _deprecated_default_inputs(self):
        """
        Handle migration of deprecated `default_inputs` attribute to `_ewoks_default_inputs`.
        """
        adict = dict(self.default_inputs)
        if not adict:
            return
        self.default_inputs.clear()
        adict = {
            name: value
            for name, value in adict.items()
            if not invalid_data.is_invalid_data(value)
            and name not in self._ewoks_default_inputs
        }
        warnings.warn(
            ".ows file node property 'default_inputs' has been converted to '_ewoks_default_inputs'. Please save the workflow to keep this change.",
            DeprecationWarning,
        )
        self.update_default_inputs(**adict)

    # --- Ewoks task dynamic inputs (NOT SAVED IN FILE) --------------------------------------------------------------

    def get_dynamic_input_names(self, include_missing: bool = False) -> set:
        """
        Return input names that have dynamic variables (or all input names).

        :param include_missing: If True return all defined input names.
        :return: Set of input names.
        """
        if include_missing:
            return set(self.get_input_names())
        else:
            return set(self.__dynamic_inputs)

    def get_dynamic_input_values(
        self, include_missing: bool = False, defaults: Optional[Mapping] = None
    ) -> dict:
        """
        Return dynamic input values or a mapping including missing markers.

        :param include_missing: If True include all input names set to INVALIDATION_DATA initially.
        :param defaults: Optional mapping of default overrides.
        :return: Dict of input name -> value or missing marker.
        """
        if include_missing:
            values = {
                name: invalid_data.INVALIDATION_DATA for name in self.get_input_names()
            }
        else:
            values = dict()
        if defaults:
            values.update(defaults)
        values.update(
            {k: self._extract_value(v) for k, v in self.__dynamic_inputs.items()}
        )
        return {name: invalid_data.as_missing(value) for name, value in values.items()}

    def get_dynamic_input_value(self, name: str, default: Any = None) -> Any:
        """
        Get a dynamic input value provided by upstream nodes.

        :param name: Input name.
        :param default: Fallback if not present.
        :return: The dynamic value or provided fallback.
        """
        value = self.__dynamic_inputs.get(name, default)
        return self._extract_value(value)

    def set_dynamic_input(self, name: str, value: Any) -> None:
        """
        Set or remove a dynamic input variable (from upstream nodes).

        :param name: Input name.
        :param value: Input variable or value. Invalid data removes the dynamic input.
        """
        if invalid_data.is_invalid_data(value):
            _logger.debug("ewoks widget: remove dynamic input %r", name)
            _ = self.__dynamic_inputs.pop(name, None)
        else:
            _logger.debug(
                "ewoks widget: set dynamic input %r = %s",
                name,
                value_from_transfer(value, varinfo=self._ewoks_varinfo),
            )
            self.__dynamic_inputs[name] = value

    def update_dynamic_inputs(self, **inputs) -> None:
        """
        Batch-set dynamic inputs.

        :param inputs: name=value pairs to set as dynamic inputs.
        """
        for name, value in inputs.items():
            self.set_dynamic_input(name, value)

    def _extract_value(self, data) -> Any:
        """
        Convert transfer objects (Variable wrappers or raw values) to plain values.

        :param data: The transferred data.
        :return: Extracted underlying value.
        """
        return value_from_transfer(data, varinfo=self._ewoks_varinfo)

    def _receive_dynamic_input(self, name: str, value: Any) -> None:
        """
        Deprecated alias for setting a dynamic input.

        :param name: Input name.
        :param value: Input value.
        """
        warnings.warn(
            "`_receive_dynamic_input` is deprecated in favor of `set_dynamic_input`.",
            DeprecationWarning,
        )
        self.set_dynamic_input(name, value)

    # --- Ewoks task outputs --------------------------------------------------------------

    @classmethod
    def get_output_names(cls):
        """
        Return Ewoks task output names for the bound task class.

        :return: Iterable of output name strings.
        """
        return cls.ewokstaskclass.output_names()

    def get_task_outputs(self) -> dict:
        """
        Return task output variables.

        Subclasses must implement this to return a dict-like mapping of output name
        to Variable.
        """
        raise NotImplementedError("Base class")

    def get_task_output_values(self) -> dict:
        """
        Return all task output values extracted from Variables.

        :return: Dict of output name -> plain value (missing replaced).
        """
        return {k: self._extract_value(v) for k, v in self.get_task_outputs().items()}

    def get_task_output_value(
        self, name, default: Any = missing_data.MISSING_DATA
    ) -> Any:
        """
        Retrieve a single task output value by name, returning default if missing.

        :param name: Output name.
        :param default: Fallback when missing.
        :return: The extracted output value or default.
        """
        adict = self.get_task_outputs()
        try:
            value = adict[name]
        except KeyError:
            return default
        value = self._extract_value(value)
        if missing_data.is_missing_data(value):
            return default
        return value

    # --- Upstream and downstream signals --------------------------------------------------------------

    def handleNewSignals(self) -> None:
        """
        Called by Orange after all signal handlers have run to set dynamic inputs.

        Default implementation triggers task execution (with propagation).
        """
        self.execute_ewoks_task(log_missing_inputs=False)

    def propagate_downstream(self, succeeded: Optional[bool] = None) -> None:
        """
        Trigger downstream propagation: send outputs on success or invalidation on failure.

        :param succeeded: Optional override of the current task success flag.
        """
        if succeeded is None:
            succeeded = self.task_succeeded
        if succeeded:
            self.__post_task_execute([self.trigger_downstream])
        else:
            self.__post_task_execute([self.clear_downstream])

    def trigger_downstream(self) -> None:
        """
        Send the current task output variables downstream via Orange signals.

        Outputs set to invalidation data are sent as INVALIDATION_DATA.
        """
        _logger.debug("%s: trigger downstream", self)
        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            for ewoksname, var in self.get_task_outputs().items():
                output = self._get_output_signal(ewoksname)
                if invalid_data.is_invalid_data(var.value):
                    self.send(output.name, invalid_data.INVALIDATION_DATA)
                    # Note: perhaps `self.invalidate(output.name)` is equivalent
                else:
                    self.send(output.name, var)
        else:
            for ewoksname, var in self.get_task_outputs().items():
                output = self._get_output_signal(ewoksname)
                if invalid_data.is_invalid_data(var.value):
                    output.send(invalid_data.INVALIDATION_DATA)
                    # Note: perhaps `output.invalidate()` is equivalent
                else:
                    output.send(var)

    def clear_downstream(self) -> None:
        """
        Propagate INVALIDATION_DATA to all downstream outputs.

        Useful to indicate that this node's outputs are invalid (e.g., after failure).
        """
        _logger.debug("%s: clear downstream", self)
        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            for ewoksname in self.get_task_outputs():
                output = self._get_output_signal(ewoksname)
                self.send(output.name, invalid_data.INVALIDATION_DATA)
                # Note: perhaps `self.invalidate(output.name)` is equivalent
        else:
            for ewoksname in self.get_task_outputs():
                output = self._get_output_signal(ewoksname)
                output.send(invalid_data.INVALIDATION_DATA)
                # Note: perhaps `output.invalidate` is equivalent

    def _get_output_signal(self, ewoksname: str) -> Output:
        """
        Resolve and return the Orange output signal for a given Ewoks output name.

        :param ewoksname: Ewoks output name.
        :raises RuntimeError: If the corresponding Orange output signal does not exist.
        :return: The Orange signal object.
        """
        return get_signal(self, "outputs", ewoksname)

    # --- Ewoks task execution --------------------------------------------------------------

    @property
    def task_output_changed_callbacks(self) -> list:
        """
        Access the list of callbacks executed after task output change.

        :return: List of callables.
        """
        return self.__task_output_changed_callbacks

    def task_output_changed(self) -> None:
        """
        Default callback invoked when task output changed.

        Subclasses may override to react to this event.
        """
        pass

    def execute_ewoks_task(self, log_missing_inputs: bool = True) -> None:
        """
        Execute the Ewoks task and propagate downstream on completion.

        :param log_missing_inputs: Whether missing inputs should be logged.
        """
        _logger.debug("%s: execute ewoks task (with propagation)", self)
        self._execute_ewoks_task(propagate=True, log_missing_inputs=log_missing_inputs)

    def execute_ewoks_task_without_propagation(self) -> None:
        """
        Execute the Ewoks task without propagating outputs downstream.
        """
        _logger.debug("%s: execute ewoks task (without propagation)", self)
        self._execute_ewoks_task(propagate=False, log_missing_inputs=False)

    @property
    def task_succeeded(self) -> Optional[bool]:
        """
        Whether the most recent task execution succeeded.

        :return: True if succeeded, False if failed, or None if never run.
        """
        raise NotImplementedError("Base class")

    @property
    def task_done(self) -> Optional[bool]:
        """
        Whether the most recent task execution finished (success or failure).

        :return: True/False or None if never run.
        """
        raise NotImplementedError("Base class")

    @property
    def task_exception(self) -> Optional[Exception]:
        """
        Exception raised during the most recent task execution, if any.

        :return: Exception instance or None.
        """
        raise NotImplementedError("Base class")

    @property
    def post_task_exception(self) -> Optional[Exception]:
        """
        Exception raised while running post-task callbacks (if any).

        :return: Exception instance or None.
        """
        return self.__post_task_exception

    def _get_task_arguments(self) -> dict:
        """
        Build task constructor arguments.

        :return: Dict with inputs, varinfo, execinfo and node_id suitable for Task constructor.
        """
        if self.signalManager is None:
            execinfo = None
            node_id = None
        else:
            scheme = self.signalManager.scheme()
            node = scheme.node_for_widget(self)
            node_id = node.title
            if not node_id:
                node_id = scheme.nodes.index(node)
            execinfo = scheme_ewoks_events(scheme, self._ewoks_execinfo)

        if self._ewoks_task_options:
            task_arguments = dict(self._ewoks_task_options)
        else:
            task_arguments = dict()
        task_arguments.update(
            inputs=self.get_task_inputs(),
            varinfo=self._ewoks_varinfo,
            execinfo=execinfo,
            node_id=node_id,
        )
        return task_arguments

    def _output_changed(self) -> None:
        """
        Called when the Ewoks task execution finishes and outputs changed.

        This base class does not call it. It is up to the derived classes
        that implement `_execute_ewoks_task` to call it.

        This invokes registered post-task callbacks.
        """
        self.__post_task_execute(self.__task_output_changed_callbacks)

    def __post_task_execute(self, callbacks: List[Callable[[], None]]) -> None:
        """
        Execute a list of callbacks sequentially.

        If a callback raises, it is stored in :attr:`__post_task_exception` and re-raised.

        :param callbacks: List of zero-argument callables to invoke.
        """
        ncallbacks = len(callbacks)
        if ncallbacks == 0:
            return
        try:
            callbacks[0]()
        except Exception as e:
            self.__post_task_exception = e
            raise
        finally:
            if ncallbacks > 1:
                self.__post_task_execute(callbacks[1:])

    def _execute_ewoks_task(self, propagate: bool, log_missing_inputs: bool) -> None:
        """
        Subclasses must implement how the task is created and executed.

        :param propagate: Whether to propagate outputs downstream after execution.
        :param log_missing_inputs: Whether to log missing input warnings.
        """
        raise NotImplementedError("Base class")



import warnings

from ..gui.qt_utils.app import QtEvent  # noqa F401
from ..gui.qt_utils.app import close_qtapp  # noqa F401
from ..gui.qt_utils.app import ensure_qtapp  # noqa F401
from ..gui.qt_utils.app import get_all_qtwidgets  # noqa F401
from ..gui.qt_utils.app import get_qtapp  # noqa F401
from ..gui.qt_utils.app import process_qtapp_events  # noqa F401
from ..gui.qt_utils.app import qt_message_handler  # noqa F401
from ..gui.qt_utils.app import qtapp_context  # noqa F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.qt_utils.qtapp' module.",
    DeprecationWarning,
    stacklevel=2,
)



from contextlib import ExitStack

from ewokscore import events
from ewokscore.events.contexts import ExecInfoType
from ewokscore.events.contexts import RawExecInfoType


def scheme_ewoks_events(scheme, execinfo: RawExecInfoType = None) -> ExecInfoType:
    scheme_execinfo = getattr(scheme, "ewoks_execinfo", None)
    if scheme_execinfo is not None:
        return scheme_execinfo
    exitstack = ExitStack()
    stack = exitstack.__enter__()
    ctx = events.job_context(execinfo)
    execinfo = stack.enter_context(ctx)
    ctx = events.workflow_context(execinfo, workflow=scheme.title)
    execinfo = stack.enter_context(ctx)
    scheme.ewoks_execinfo = execinfo

    def ewoks_finalize():
        # TODO: job and workflow end event will never capture
        #       node exceptions because they are absorbed by orange.
        exitstack.close()

    scheme.ewoks_finalize = ewoks_finalize
    scheme.destroyed.connect(ewoks_finalize)
    return execinfo



from typing import List

from .pkg_meta import get_distribution


def widget_discovery(discovery, distroname: str, modules: List[str]):
    dist = get_distribution(distroname)
    for widget_category_module in modules:
        discovery.process_category_package(widget_category_module, distribution=dist)



import gc
import logging
import signal
import sys
import time
from contextlib import contextmanager
from typing import Iterator
from typing import Optional

from AnyQt import QtCore
from AnyQt.QtWidgets import QApplication

_APP = None
_OLD_HANDLERS = None

logger = logging.getLogger(__name__)


def ensure_qtapp() -> Optional[QApplication]:
    """Create a Qt application without event loop when no application is running.
    Returns None when the application already exists."""
    global _APP
    if _APP is not None:
        return

    # GUI application
    _APP = get_qtapp()
    if _APP is not None:
        return

    # Install signal, exception and Qt message handlers
    _install_handlers()

    # Application without event loop (_APP.exec() is not called)
    _APP = QApplication([])
    return _APP


def _install_handlers() -> None:
    """Install signal, exception and Qt message handlers"""
    global _OLD_HANDLERS
    if _OLD_HANDLERS is not None:
        return
    # Allow termination with CTRL + C
    old_signal = signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Redirect Qt logs
    old_qtmsg_handler = QtCore.qInstallMessageHandler(qt_message_handler)

    # Log unhandled exceptions, otherwise qFatal will be called
    old_ex_handler = sys.excepthook
    sys.excepthook = absorb_nonbase_exception

    _OLD_HANDLERS = old_signal, old_qtmsg_handler, old_ex_handler


def _remove_handlers() -> None:
    """Undo _install_handlers"""
    global _OLD_HANDLERS
    if _OLD_HANDLERS is None:
        return

    old_signal, old_qtmsg_handler, old_ex_handler = _OLD_HANDLERS

    signal.signal(signal.SIGINT, old_signal)

    QtCore.qInstallMessageHandler(old_qtmsg_handler)

    sys.excepthook = old_ex_handler

    _OLD_HANDLERS = None


def close_qtapp() -> None:
    """Close the Qt application created by ensure_qtapp"""
    global _APP
    if _APP is None:
        return
    _APP.processEvents()
    while gc.collect():
        _APP.processEvents()
    _APP.exit()
    _APP = None
    _remove_handlers()


@contextmanager
def qtapp_context() -> Iterator[Optional[QApplication]]:
    """Yields None when the Qt application already exists"""
    qtapp = ensure_qtapp()
    try:
        yield qtapp
    finally:
        if qtapp is not None:
            close_qtapp()


def get_qtapp() -> Optional[QApplication]:
    """Get the QApplication in the current process (if any)."""
    return QApplication.instance()


def process_qtapp_events() -> None:
    """Process all pending Qt events when a Qt event loop is running"""
    if _APP is None:
        return
    _APP.processEvents(QtCore.QEventLoop.AllEvents)


class QtEvent:
    """Event that also works for Qt applications with an event loop
    that needs to run manually"""

    def __init__(self) -> None:
        self.__flag = False

    def wait(self, timeout: Optional[float] = None) -> bool:
        """Processes events associated to the calling thread while waiting"""
        if timeout is not None:
            t0 = time.time()
        while not self.__flag:
            if _APP is None:
                time.sleep(0.1)
            else:
                _APP.processEvents(QtCore.QEventLoop.AllEvents, 100)
            if timeout is not None:
                if (time.time() - t0) > timeout:
                    return False
        return True

    def set(self) -> None:
        self.__flag = True

    def clear(self) -> None:
        self.__flag = False


def get_all_qtwidgets() -> list:
    app = get_qtapp()
    if app is None:
        return list()

    sapp = str(type(app))
    if "PyQt6" in sapp:
        from PyQt6.sip import ispycreated as createdByPython  # noqa
    elif "PyQt5" in sapp or "PyQt6" in sapp:
        from PyQt5.sip import ispycreated as createdByPython  # noqa
    elif "PySide2" in sapp:
        from PySide2.shiboken2 import createdByPython  # noqa
    else:
        raise RuntimeError(f"'{sapp}' not supported")

    return [widget for widget in app.allWidgets() if createdByPython(widget)]


def qt_message_handler(level, context, message) -> None:
    if level == QtCore.QtInfoMsg:
        log = logger.info
    elif level == QtCore.QtWarningMsg:
        log = logger.warning
    elif level == QtCore.QtCriticalMsg:
        log = logger.error
    elif level == QtCore.QtFatalMsg:
        log = logger.fatal
    else:
        log = logger.debug
    log(
        "line: %d, func: %s(), file: %s\nQT MESSAGE: %s\n",
        context.line,
        context.function,
        context.file,
        message,
    )


def absorb_nonbase_exception(exc_type, exc_value, exc_traceback) -> None:
    if not issubclass(exc_type, Exception):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))



import warnings

from ..gui.canvas.config import Config  # noqa F401
from ..gui.canvas.config import widgets_entry_points  # noqa F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.canvas.config' module.",
    DeprecationWarning,
    stacklevel=2,
)



import warnings

from ewoksorange.gui.orange_utils.settings import Setting  # noqa: F401
from ewoksorange.gui.orange_utils.settings import get_settings  # noqa: F401
from ewoksorange.gui.orange_utils.settings import is_setting  # noqa: F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.orange_utils.owsettings' module.",
    DeprecationWarning,
    stacklevel=2,
)



import warnings

from ..gui.workflows.task_wrappers import OWWIDGET_TASKS_GENERATOR

__all__ = ["OWWIDGET_TASKS_GENERATOR"]

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.bridge.task_wrappers' module.",
    DeprecationWarning,
    stacklevel=2,
)



from ewokscore.tests.examples.tasks.sumtask import SumTask

from ..gui.orange_utils.signals import Input
from ..gui.orange_utils.signals import Output
from ..gui.owwidgets.nothread import OWEwoksWidgetNoThread
from .utils import execute_task


class CustomSignalsWidget(OWEwoksWidgetNoThread, ewokstaskclass=SumTask):
    name = "custom_signals"

    class Inputs:
        c = Input("A", object, ewoksname="a")
        d = Input("B", object, ewoksname="b")

    class Outputs:
        e = Output("A + B", object, ewoksname="result")


def test_execute_custom_signals_widget(qtapp):
    result = execute_task(CustomSignalsWidget, inputs={"a": 5, "b": 6})
    expected = {"result": 11}
    assert result == expected, result



import pathlib
from typing import Any
from typing import List
from typing import Optional
from typing import Union

from ewokscore.engine_interface import Path
from ewokscore.engine_interface import RawExecInfoType
from ewokscore.engine_interface import TaskGraph
from ewokscore.engine_interface import WorkflowEngineWithSerialization

# Note: lazy import of bindings because it relies on Qt and
# we do not want to import Qt when not needed because the Qt binding
# is an optional dependency.


class OrangeWorkflowEngine(WorkflowEngineWithSerialization):

    def execute_graph(
        self,
        graph: Any,
        *,
        inputs: Optional[List[dict]] = None,
        load_options: Optional[dict] = None,
        varinfo: Optional[dict] = None,
        execinfo: RawExecInfoType = None,
        task_options: Optional[dict] = None,
        outputs: Optional[List[dict]] = None,
        merge_outputs: Optional[bool] = True,
        # Engine specific:
        error_on_duplicates: bool = True,
        tmpdir: Optional[str] = None,
    ) -> None:
        from .bindings import execute_graph

        execute_graph(
            graph,
            inputs=inputs,
            load_options=load_options,
            varinfo=varinfo,
            execinfo=execinfo,
            task_options=task_options,
            outputs=outputs,
            merge_outputs=merge_outputs,
            error_on_duplicates=error_on_duplicates,
            tmpdir=tmpdir,
        )

    def deserialize_graph(
        self,
        graph: Any,
        *,
        inputs: Optional[List[dict]] = None,
        representation: Optional[str] = None,
        root_dir: Optional[Union[str, Path]] = None,
        root_module: Optional[str] = None,
        # Serializer specific:
        preserve_ows_info: Optional[bool] = True,
        title_as_node_id: Optional[bool] = False,
    ) -> TaskGraph:
        from .bindings import load_graph

        return load_graph(
            graph,
            inputs=inputs,
            representation=representation,
            root_dir=root_dir,
            root_module=root_module,
            preserve_ows_info=preserve_ows_info,
            title_as_node_id=title_as_node_id,
        )

    def serialize_graph(
        self,
        graph: TaskGraph,
        destination,
        *,
        representation: Optional[str] = None,
        # Serializer specific:
        **serialize_options,
    ) -> Union[str, dict]:
        from .bindings import save_graph

        return save_graph(
            graph, destination, representation=representation, **serialize_options
        )

    def get_graph_representation(self, graph: Any) -> Optional[str]:
        if isinstance(graph, str) and graph.endswith(".ows"):
            return "ows"
        if isinstance(graph, pathlib.Path) and graph.suffix == ".ows":
            return "ows"



import importlib.metadata
from types import MethodType

from packaging.version import Version

from ..orange_version import ORANGE_VERSION
from ..pkg_meta import get_distribution


def oasys_patch():
    """OASYS1 and ewoksorange have conflicting dependencies.
    This patch ensures the oasys.widgets entry points can
    be resolved (missing dependencies causes them to fail).
    """
    if ORANGE_VERSION != ORANGE_VERSION.oasys_fork:
        return

    if Version(importlib.metadata.version("oasys-canvas-core")) >= Version("1.0.10"):
        return

    def requires(self, extras=()):
        return []

    dist = get_distribution("OASYS1", raise_error=True)
    dist.requires = MethodType(requires, dist)
    dist = get_distribution("ewoksorange", raise_error=True)
    dist.requires = MethodType(requires, dist)



import warnings

from ..gui.workflows.owscheme import ewoks_to_ows  # noqa: F401
from ..gui.workflows.owscheme import graph_is_supported  # noqa: F401
from ..gui.workflows.owscheme import ows_to_ewoks  # noqa: F401
from ..gui.workflows.owscheme import patch_parse_ows_stream  # noqa: F401

__all__ = ["ows_to_ewoks", "ewoks_to_ows", "graph_is_supported"]

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.bridge.owscheme' module.",
    DeprecationWarning,
    stacklevel=2,
)



import importlib.metadata
from enum import Enum

_OrangeVersion = Enum("OrangeVersion", "latest_orange oasys_fork latest_orange_base")

_DISTRIBUTION_TO_VERSION = {
    "oasys1": _OrangeVersion.oasys_fork,
    "oasys-canvas-core": _OrangeVersion.oasys_fork,
    "orange3": _OrangeVersion.latest_orange,
    "orange-canvas-core": _OrangeVersion.latest_orange_base,
}

for distribution, version in _DISTRIBUTION_TO_VERSION.items():
    try:
        _ = importlib.metadata.version(distribution)
    except importlib.metadata.PackageNotFoundError:
        continue
    else:
        ORANGE_VERSION = version
        break
else:
    raise importlib.metadata.PackageNotFoundError(
        "No compatible Orange distributions found."
    )



from contextlib import ExitStack
from contextlib import contextmanager
from typing import Generator

from AnyQt import QtWidgets


@contextmanager
def block_signals(*widgets: QtWidgets.QWidget) -> Generator[None, None, None]:
    """
    Context manager that blocks signals on one or more Qt widgets.

    .. code-block:: python

        with block_signals(widget1, widget2):
            ...
    """
    with ExitStack() as stack:
        for w in widgets:
            old: bool = w.blockSignals(True)
            stack.callback(w.blockSignals, old)
        yield



from ewokscore import Task

from ..gui.owwidgets.meta import ow_build_opts
from ..gui.owwidgets.threaded import (
    OWEwoksWidgetOneThreadPerRun as _OWEwoksWidgetOneThreadPerRun,
)
from ..gui.qt_utils.app import QtEvent


class MyObject:
    def __init__(self):
        self.value = None
        self.finished = QtEvent()

    def finished_callback(self):
        self.finished.set()


class DummyTask(
    Task,
    input_names=("my_object", "value"),
    output_names=("my_object",),
):
    """Task that set a value to MyObject and set a 'finished' Event"""

    def run(self):
        my_object = self.inputs.my_object
        my_object.value = self.inputs.value
        self.outputs.my_object = my_object
        my_object.finished_callback()


class OWEwoksWidgetOneThreadPerRun(
    _OWEwoksWidgetOneThreadPerRun,
    **ow_build_opts,
    ewokstaskclass=DummyTask,
):
    name = "test_OW"


def test_OWEwoksWidgetOneThreadPerRun(qtapp):
    """
    Test processing several tasks.
    The widget will create one thread per task and execution will be done in parallel.
    Make sure all tasks are completed with valid outputs.
    """
    widget = OWEwoksWidgetOneThreadPerRun()

    objects = (
        MyObject(),
        MyObject(),
        MyObject(),
    )

    for value, obj in enumerate(objects):
        widget.set_dynamic_input("value", value)
        widget.set_dynamic_input("my_object", obj)

        # Start calculation
        widget.handleNewSignals()

    for obj in objects:
        obj.finished.wait(timeout=3)

    values = [obj.value for obj in objects]
    expected = [0, 1, 2]
    assert values == expected



import pytest

from ..gui.owwidgets.base import OWWidget
from ..orange_version import ORANGE_VERSION
from .utils import execute_task

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:

    class NativeOldStyleWidget(OWWidget):
        name = "native"
        inputs = [("A", object, "set_a"), ("B", object, "set_b")]
        outputs = [{"name": "A + B", "id": "A + B", "type": object}]

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._a = 0
            self._b = 0

        def set_a(self, value):
            self._a = value
            self.send("A + B", self._a + self._b)

        def set_b(self, value):
            self._b = value
            self.send("A + B", self._a + self._b)

    NativeNewStyleWidget = NotImplemented
else:
    from orangewidget.widget import Input
    from orangewidget.widget import Output

    class NativeNewStyleWidget(OWWidget):
        name = "native"

        class Inputs:
            a = Input("A", object)
            b = Input("B", object)

        class Outputs:
            result = Output("A + B", object)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._a = 0
            self._b = 0

        @Inputs.a
        def set_a(self, value):
            self._a = value
            self.Outputs.result.send(self._a + self._b)

        @Inputs.b
        def set_b(self, value):
            self._b = value
            self.Outputs.result.send(self._a + self._b)

    class NativeOldStyleWidget(OWWidget):
        name = "native"
        inputs = [("A", object, "set_a"), ("B", object, "set_b")]
        outputs = [("A + B", object)]  # dict does not work

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._a = 0
            self._b = 0

        def set_a(self, value):
            self._a = value
            self.send("A + B", self._a + self._b)

        def set_b(self, value):
            self._b = value
            self.send("A + B", self._a + self._b)


@pytest.mark.parametrize("widget_class", [NativeOldStyleWidget, NativeNewStyleWidget])
def test_execute_native_widget(qtapp, widget_class):
    if widget_class is NotImplemented:
        pytest.skip(f"Not supported by {ORANGE_VERSION}")
    if widget_class is NativeOldStyleWidget:
        inputs = {"A": 5, "B": 6}
        expected = {"A + B": 11}
    else:
        inputs = {"a": 5, "b": 6}
        expected = {"result": 11}
    result = execute_task(widget_class, inputs=inputs)
    assert result == expected, result


def test_execute_python_script(qtapp):
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        from oasys.widgets.tools.ow_python_script import OWPythonScript
        from oasys.widgets.tools.ow_python_script import Script

        settings = {
            "auto_execute": True,
            "currentScriptIndex": 0,
            "libraryListSource": [Script("test_script", "out_object = [1, 2, 3]")],
        }
        expected = {"out_object": [1, 2, 3]}
    elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        from Orange.widgets.data.owpythonscript import OWPythonScript

        settings = {
            "scriptText": "out_object = [1, 2, 3]",
            "scriptLibrary": [],
            "__version__": 2,
        }
        expected = {
            "data": None,
            "learner": None,
            "classifier": None,
            "object": [1, 2, 3],
        }
    else:
        pytest.skip("Requires the Orange3 or Oasys1 python script widget")

    result = execute_task(OWPythonScript, settings=settings)
    assert result == expected, result



from typing import List
from typing import Optional
from typing import Tuple

from ewokscore.model import BaseInputModel
from ewokscore.task import Task
from pydantic import Field

from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ewoksorange.gui.owwidgets.registration import _temporary_widget_discovery_object
from ewoksorange.gui.owwidgets.registration import register_owwidget


class InputModel(BaseInputModel):
    a: int = Field(default=12)
    b: Tuple[str] = Field(
        default=("b",),
    )
    c: List[float]
    d: Optional[float] = Field(default=None)


class TaskA(
    Task,
    input_model=InputModel,
):
    def run(self):
        pass


class EwoksOrangeTaskA(OWEwoksWidgetNoThread, ewokstaskclass=TaskA):
    name = "ewoks widget A"


def test_default_values(ewoks_orange_canvas):
    """
    Test that task with an ewoks InputModel are taking into account field default values and default factory.

    Warning: Orange consider "None" as the value to 'invalidate' link. As a consequence those are filtered if defined as a default value.
    """
    widget_registry = _temporary_widget_discovery_object()

    for widget in (EwoksOrangeTaskA,):
        register_owwidget(
            widget_class=widget,
            package_name="ewoksorange",
            category_name="test",
            project_name="ewoksorange",
            discovery_object=widget_registry,
        )

    widget_a = EwoksOrangeTaskA()
    assert widget_a.get_default_input_values() == {
        "a": 12,
        "b": ("b",),
    }



import warnings

from ..gui.orange_utils.signal_manager import (  # noqa F401
    SignalManagerWithOutputTracking,
)
from ..gui.orange_utils.signal_manager import SignalManagerWithoutScheme  # noqa F401
from ..gui.orange_utils.signal_manager import SignalManagerWithScheme  # noqa F401
from ..gui.orange_utils.signal_manager import patch_signal_manager  # noqa F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui...' module.",
    DeprecationWarning,
    stacklevel=2,
)



import warnings

__all__ = [
    "convert_graph",
    "execute_graph",
    "graph_is_supported",
    "load_graph",
    "save_graph",
]


def __getattr__(name):
    """Lazy import with deprecation warning."""
    if name not in __all__:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    from . import bindings

    warnings.warn(
        f"Accessing '{name}' from '{__name__}' is deprecated and will be "
        "removed in a future release. Please import it from "
        "'ewoksorange.bindings' instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(bindings, name)



import warnings

__all__ = ["execute_graph", "load_graph", "save_graph", "convert_graph"]

from .. import convert_graph
from .. import execute_graph
from .. import load_graph
from .. import save_graph

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.bindings' module.",
    DeprecationWarning,
    stacklevel=2,
)



"""
Metaclass and class preparation utilities for owwidgets package.
"""

import inspect
from typing import Any

from ...orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from orangewidget.settings import Setting
    from orangewidget.widget import WidgetMetaClass
else:
    from orangewidget.settings import Setting

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        from Orange.widgets.widget import WidgetMetaClass
    else:
        from orangewidget.widget import OWBaseWidget

        WidgetMetaClass = type(OWBaseWidget)

from ..orange_utils import _signals


class OWEwoksWidgetMetaClass(WidgetMetaClass):
    """
    Metaclass used to prepare widget classes with Ewoks bindings.
    """

    def __new__(metacls, name, bases, attrs, ewokstaskclass=None, **kw):
        """
        Create a new widget class; if `ewokstaskclass` is provided prepare the class.

        :param name: New class name.
        :param bases: Base classes.
        :param attrs: Attribute dict for the class.
        :param ewokstaskclass: Optional Ewoks Task class to attach.
        :return: Newly created class type.
        """
        if ewokstaskclass:
            _prepare_OWEwoksWidgetclass(attrs, ewokstaskclass)
        return super().__new__(metacls, name, bases, attrs, **kw)


# Ensure compatibility between old orange widget and new
# orangewidget.widget.WidgetMetaClass. This was before the split of the two
# projects. Parameter name "openclass" is undefined in the old version.
ow_build_opts = dict()
if "openclass" in inspect.signature(WidgetMetaClass).parameters:
    ow_build_opts["openclass"] = True


def _prepare_OWEwoksWidgetclass(namespace: dict, ewokstaskclass: Any) -> None:
    """
    Attach Ewoks task class and default settings to a widget class namespace.
    This needs to be called before signal and setting parsing.

    :param namespace: Class attribute dictionary to modify.
    :param ewokstaskclass: The Ewoks Task class to attach (used for input/output introspection).
    """

    # Add the Ewoks class as an attribute to the Orange widget class
    namespace["ewokstaskclass"] = ewokstaskclass

    # Make sure the values above are always the default setting values:
    # https://orange3.readthedocs.io/projects/orange-development/en/latest/tutorial-settings.html
    # schema_only=False: when a widget is removed, its settings are stored to be used
    #                    as defaults for future instances of this widget.
    # schema_only=True: setting defaults should not change. Future instances of this widget
    #                   have the default settings hard-coded in this function.
    schema_only = True

    # Add the settings as widget class attributes
    namespace["_ewoks_default_inputs"] = Setting(dict(), schema_only=schema_only)
    namespace["_ewoks_varinfo"] = Setting(dict(), schema_only=schema_only)
    namespace["_ewoks_execinfo"] = Setting(dict(), schema_only=schema_only)
    namespace["_ewoks_task_options"] = Setting(dict(), schema_only=schema_only)

    # Deprecated:
    namespace["default_inputs"] = Setting(dict(), schema_only=schema_only)

    # Add missing inputs and outputs as widget class attributes
    _signals.validate_signals(
        namespace,
        "inputs",
        name_to_ignore=namespace.get("_ewoks_inputs_to_hide_from_orange", tuple()),
    )
    _signals.validate_signals(
        namespace,
        "outputs",
        name_to_ignore=namespace.get("_ewoks_outputs_to_hide_from_orange", tuple()),
    )



import warnings

from .gui.owwidgets.registration import get_owwidget_descriptions  # noqa F401
from .gui.owwidgets.registration import register_owwidget  # noqa F401
from .gui.owwidgets.registration import widget_discovery  # noqa F401

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.owwidgets.registration' module.",
    DeprecationWarning,
    stacklevel=2,
)



import weakref
from typing import Any

from ...orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    import AnyQt.QtCore  # noqa: F401 isort: skip  Needed for "import sip"
    import oasys.canvas.widgetsscheme as widgetsscheme_module
    from oasys.canvas.widgetsscheme import (
        OASYSSignalManager as _SignalManagerWithSchemeOrg,
    )

    class _SignalManagerWithScheme(_SignalManagerWithSchemeOrg):
        def has_pending(self):
            return bool(self._input_queue)

    notify_input_helper = None
else:
    from orangewidget.workflow.widgetsscheme import (
        WidgetsSignalManager as _SignalManagerWithScheme,
    )
    import orangewidget.workflow.widgetsscheme as widgetsscheme_module
    from orangewidget.utils.signals import notify_input_helper

from ewokscore.variable import Variable
from ewokscore.variable import value_from_transfer
from orangecanvas.scheme import signalmanager

from ..owwidgets.types import is_native_widget
from ..qt_utils.app import QtEvent
from ..utils import invalid_data
from ._signals import get_signal_orange_names

# monkey patch of 'can_enable_dynamic' See https://gitlab.esrf.fr/workflow/ewoks/ewoksorange/-/issues/58

_super_can_enable_dynamic = signalmanager.can_enable_dynamic


def can_enable_dynamic_patch(link, value):
    # type: (SchemeLink, Any) -> bool
    """
    Can the a dynamic `link` (:class:`SchemeLink`) be enabled for `value`.
    """
    if isinstance(value, Variable):
        value = value.value
    return _super_can_enable_dynamic(link, value)


signalmanager.can_enable_dynamic = can_enable_dynamic_patch

# end monkey patch


class _MissingSignalValue:
    """Indicates a missing signal value and supports waiting for the real signal value"""

    completed = QtEvent()


class _OwWidgetSignalValues:
    """Store signal values and support waiting for a value"""

    def __init__(self):
        self._values = dict()

    def _get_value(self, signal_name) -> Any:
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        if signal_name not in self._values:
            self._values[signal_name] = _MissingSignalValue()
        return self._values[signal_name]

    def _set_value(self, signal_name, value) -> None:
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        self._values[signal_name] = value

    def set_value(self, signal_name, value) -> None:
        previous_value = self._get_value(signal_name)
        self._set_value(signal_name, value)
        if isinstance(previous_value, _MissingSignalValue):
            previous_value.completed.set()

    def invalidate_value(self, signal_name) -> None:
        previous_value = self._get_value(signal_name)
        if isinstance(previous_value, _MissingSignalValue):
            return
        self.set_value(signal_name, _MissingSignalValue())

    def get_value(self, signal_name, timeout=None) -> Any:
        value = self._get_value(signal_name)
        if isinstance(value, _MissingSignalValue):
            value.completed.wait(timeout=timeout)
            value = self._get_value(signal_name)
        return value

    def has_value(self, signal_name) -> bool:
        value = self._get_value(signal_name)
        return not isinstance(value, _MissingSignalValue)

    def has_values(self) -> bool:
        return bool(self._values)


class _OwWidgetSignalState:
    """Store input and output signal values for one widget and support waiting for a value"""

    def __init__(self, owwidget, *args, **kwargs):
        output_variable_names = get_signal_orange_names(owwidget, "outputs")
        widget_has_outputs = bool(output_variable_names)
        if not output_variable_names:
            output_variable_names = get_signal_orange_names(owwidget, "inputs")

        self._variable_names = output_variable_names
        self._widget_has_outputs = widget_has_outputs
        self._input_values = _OwWidgetSignalValues()
        self._output_values = _OwWidgetSignalValues()
        super().__init__(*args, **kwargs)

    def set_output_value(self, signal_name, value) -> None:
        self._output_values.set_value(signal_name, value)

    def invalidate_output_value(self, signal_name) -> None:
        self._output_values.invalidate_value(signal_name)

    def get_output_value(self, signal_name, timeout=None) -> Any:
        return self._output_values.get_value(signal_name, timeout=timeout)

    def has_output_value(self, signal_name) -> bool:
        return self._output_values.has_value(signal_name)

    def has_output_values(self) -> bool:
        return self._output_values.has_values()

    def set_input_value(self, signal_name, value) -> None:
        self._input_values.set_value(signal_name, value)

    def invalidate_input_value(self, signal_name) -> None:
        self._input_values.invalidate_value(signal_name)

    def get_input_value(self, signal_name, timeout=None) -> Any:
        return self._input_values.get_value(signal_name, timeout=timeout)

    def has_input_value(self, signal_name) -> bool:
        return self._input_values.has_value(signal_name)

    def has_input_values(self) -> bool:
        return self._input_values.has_values()

    def is_executed(self) -> bool:
        if self._variable_names:
            if self._widget_has_outputs:
                # Widget is executed when all outputs are set
                return all(
                    (self.has_output_value(name) for name in self._variable_names)
                )
            # Widget is executed when all inputs are set
            return all((self.has_input_value(name) for name in self._variable_names))

        # Task has no inputs or outputs
        return self.has_output_values() or self.has_input_values()


class SignalManagerWithOutputTracking:
    """Store input and output signal value per widget. Knows
    when a widget is "executed" or not.

    Losely based on Orange.widgets.tests.base.DummySignalManager
    """

    def __init__(self, *args, **kwargs):
        self._widget_states = weakref.WeakKeyDictionary()
        super().__init__(*args, **kwargs)

    def _get_widget_state(self, owwidget) -> _OwWidgetSignalState:
        state = self._widget_states.get(owwidget, None)
        if state is not None:
            return state
        state = _OwWidgetSignalState(owwidget)
        self._widget_states[owwidget] = state
        return state

    def set_output_value(self, owwidget, signal_name, value) -> None:
        self._get_widget_state(owwidget).set_output_value(signal_name, value)

    def invalidate_output_value(self, owwidget, signal_name) -> None:
        self._get_widget_state(owwidget).invalidate_output_value(signal_name)

    def get_output_value(self, owwidget, signal_name, timeout=None) -> Any:
        return self._get_widget_state(owwidget).get_output_value(
            signal_name, timeout=timeout
        )

    def has_output_value(self, owwidget, signal_name) -> bool:
        return self._get_widget_state(owwidget).has_output_value(signal_name)

    def set_input_value(self, owwidget, signal_name, value) -> None:
        self._get_widget_state(owwidget).set_input_value(signal_name, value)

    def invalidate_input_value(self, owwidget, signal_name) -> None:
        self._get_widget_state(owwidget).invalidate_input_value(signal_name)

    def get_input_value(self, owwidget, signal_name, timeout=None) -> Any:
        return self._get_widget_state(owwidget).get_input_value(
            signal_name, timeout=timeout
        )

    def has_input_value(self, owwidget, signal_name) -> bool:
        return self._get_widget_state(owwidget).has_input_value(signal_name)

    def widget_is_executed(self, owwidget) -> bool:
        return self._get_widget_state(owwidget).is_executed()


class SignalManagerWithoutScheme(SignalManagerWithOutputTracking):
    """Used when no Orange canvas is present.

    Only needs to keep track of widget outputs because data between tasks
    is passed by the Ewoks mechanism, not the Orange mechanism (signal manager).
    """

    def send(self, owwidget, signal_name, value, *args, **kwargs) -> None:
        self.set_output_value(owwidget, signal_name, value)


class SignalManagerWithScheme(
    SignalManagerWithOutputTracking, _SignalManagerWithScheme
):
    """Used when the Orange canvas is present.

    Dereference `Variable` types for native Orange widget inputs.
    """

    def send(self, owwidget, signal_name, value, *args, **kwargs) -> None:
        super().send(owwidget, signal_name, value, *args, **kwargs)
        self.set_output_value(owwidget, signal_name, value)

    def process_signals_for_widget(self, node, owwidget, signals) -> None:
        for signal in signals:
            if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
                signal_name = signal.link.sink_channel
            else:
                signal_name = signal.channel.name
            self.set_input_value(owwidget, signal_name, signal.value)
        if is_native_widget(owwidget):
            modified_signals = list()
            for signal in signals:
                sinfo = signal._asdict()
                sinfo["value"] = value_from_transfer(sinfo["value"])
                modified_signals.append(type(signal)(**sinfo))
            signals = modified_signals
        super().process_signals_for_widget(node, owwidget, signals)

    def invalidate(self, node, channel) -> None:
        super().invalidate(node, channel)
        owwidget = self.scheme().widget_for_node(node)
        if owwidget is None:
            return
        self.invalidate_input_value(owwidget, channel.name, _MissingSignalValue())

    def widget_is_finished(self, owwidget) -> bool:
        """Widget is executed for the last time and will not be execute again"""
        if self.has_pending():
            return False  # The widget might be executed again
        return super().widget_is_executed(owwidget)


def set_input_value(owwidget, signal, value, index) -> None:
    value = invalid_data.as_invalidation(value)
    key = id(owwidget), signal.name, signal.id
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        handler = getattr(owwidget, signal.handler)
        handler(value)
    else:
        notify_input_helper(signal, owwidget, value, key=key, index=index)


def patch_signal_manager():
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        widgetsscheme_module.OASYSSignalManager = SignalManagerWithScheme
    else:
        widgetsscheme_module.WidgetsSignalManager = SignalManagerWithScheme



"""Commonly used Orange3 components for implementing Orange Widgets."""

import warnings

from orangewidget.settings import Setting  # noqa F401

from ...orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets import gui  # noqa F401
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets import gui  # noqa F401
else:
    from orangewidget import gui  # noqa F401

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    _Input = None
    _Output = None
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from orangewidget.widget import Input as _Input  # noqa F401
    from orangewidget.widget import Output as _Output  # noqa F401
else:
    from orangewidget.widget import Input as _Input  # noqa F401
    from orangewidget.widget import Output as _Output  # noqa F401

# OWBaseWidget: lowest level Orange widget base class
# OWWidget: highest level Orangewidget base class.
if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets.widget import OWWidget  # noqa F401

    OWBaseWidget = OWWidget
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets.widget import OWWidget  # noqa F401
    from orangewidget.widget import OWBaseWidget  # noqa F401
else:
    from orangewidget.widget import OWBaseWidget  # noqa F401

    OWWidget = OWBaseWidget


def __getattr__(name):
    if name == "Input":
        warnings.warn(
            f"Accessing '{name}' from '{__name__}' is deprecated and will be "
            "removed in a future release. Please import it from "
            "'ewoksorange.gui.orange_utils.signals' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _Input

    if name == "Output":
        warnings.warn(
            f"Accessing '{name}' from '{__name__}' is deprecated and will be "
            "removed in a future release. Please import it from "
            "'ewoksorange.gui.orange_utils.signals' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _Output

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from orangecontrib.ewoksdemo.sumtask import OWSumTask

from .utils import execute_task


def test_sumtask_task():
    result = execute_task(OWSumTask.ewokstaskclass, inputs={"a": 1, "b": 2})
    assert result == {"result": 3}


def test_sumtask_widget(qtapp):
    result = execute_task(OWSumTask, inputs={"a": 1, "b": 2})
    assert result == {"result": 3}


def test_orange_only_input(qtapp):
    widget = OWSumTask()
    assert widget.Inputs.c



"""
Orange behavior:

- The widget instance attributes `Inputs` and `Outputs` are instances of the widget class attributes `Inputs` and `Outputs`.
- The `Input` and `Output` attributes of the `Inputs` and `Outputs` instances hold a reference to the widget.
- `Input` and `Output` attributes have two names: orange name and Inputs/Outputs container attribute name.
- Old-style `inputs` and `outputs` are deprecated but can still exist.

Ewoks-Orange behavior:

- `Input` and `Output` attributes have three names: orange name, ewoks name and Inputs/Outputs container attribute name.

Oasys behavior:

- Does not use `Inputs` or `Outputs` class, it uses lists for tuples or dicts. We create the
  `Inputs` and `Outputs` classes for ewoksorange but Oasys does not use them.

Nomenclature:

- Instances of `Input` and `Output` are referred to as "signals".

Implementation:

- When Orange instantiates a widget, it calls `OWBaseWidget._bind_signals` to sets the `Inputs` and `Outputs` attributes
  of the widget instance to be `Inputs` and `Outputs` instances. Thats why in this module we have methods that accept
  widget or signal container classes and instances:

  .. code--block:: python

    signal_container: Union[str, object]
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]]
"""

import inspect
import sys
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import get_args
from typing import get_origin

if sys.version_info >= (3, 10):
    from types import UnionType

    has_UnionType = True
else:
    has_UnionType = False
from pydantic import BaseModel

from ...orange_version import ORANGE_VERSION
from .orange_imports import OWBaseWidget
from .signals import Input
from .signals import Output
from .signals import _InputSignal
from .signals import _OutputSignal

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:

    def _native_getsignals(
        signal_container_class: type,
    ) -> Union[List[Tuple[str, Input]], List[Tuple[str, Output]]]:
        # Copied from the latest orange-widget-base
        return [
            (k, v)
            for cls in reversed(inspect.getmro(signal_container_class))
            for k, v in cls.__dict__.items()
            if isinstance(v, (Input, Output))
        ]

else:
    from orangewidget.utils.signals import getsignals as _native_getsignals


def _get_signals(
    signal_container: Union[type, object],
) -> Union[List[Tuple[str, Input]], List[Tuple[str, Output]]]:
    if isinstance(signal_container, type):
        lst = _native_getsignals(signal_container)
    else:
        lst = _native_getsignals(type(signal_container))
        lst = [(name, getattr(signal_container, name)) for name, _ in lst]
    return lst


def get_signal_list(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
) -> Union[List[Input], List[Output]]:
    """Returns list of Input or Output signal instances."""
    signal_container = _get_signal_container(orange_widget, direction)
    return _get_signal_list_from_container(signal_container)


def _get_signal_list_from_container(
    signal_container: Union[type, object],
) -> Union[List[Input], List[Output]]:
    """Returns list of Input or Output signal instances."""
    signal_list = []
    counter = 0
    for attrname, signal in _get_signals(signal_container):
        if not getattr(signal, "ewoksname", ""):
            # Most likely a native Orange/Oasys widget
            signal.ewoksname = attrname

        if getattr(signal, "_seq_id"):
            signal_list.append((signal._seq_id, signal))
        else:
            # Most likely a native Oasys widget
            counter += 1
            signal_list.append((counter, signal))

    return [signal for _, signal in sorted(signal_list, key=lambda tpl: tpl[0])]


def _get_signal_ewoks_dict(
    signal_container: Union[str, object],
) -> Dict[str, Union[List[Input], List[Output]]]:
    """Returns dict of Input or Output signal instances.
    The keys are the ewoks names or attribute names when missing.
    """
    signal_dict = {}
    for attrname, signal in _get_signals(signal_container):
        if not getattr(signal, "ewoksname", ""):
            # Most likely a native Orange widget
            signal.ewoksname = attrname
        signal_dict[signal.ewoksname] = signal
    return signal_dict


def _get_signal_container(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
) -> Union[type, object]:
    """Returns attribute which is a class with Input or Output signal instances as attributes."""
    attr_name = direction.title()

    signal_container_class = None

    if hasattr(orange_widget, direction):
        if not hasattr(orange_widget, attr_name) or not _get_signals(
            getattr(orange_widget, attr_name)
        ):
            # Most likely a native Orange/Oasys widget.
            # Old-style inputs/outputs as a list of tuples or dicts
            # instead of the new-style Inputs/Outputs classes.
            # Old-style is deprecated in Orange.
            signals = getattr(orange_widget, direction)
            signal_container_class = _oldstyle_signal_container(signals, direction)
    else:
        if not hasattr(orange_widget, attr_name):
            signal_container_class = type(attr_name, (), {})

    if signal_container_class is not None:
        if isinstance(orange_widget, type):
            setattr(orange_widget, attr_name, signal_container_class)
        else:
            setattr(orange_widget, attr_name, signal_container_class())

    return getattr(orange_widget, attr_name)


def get_signal_orange_names(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
) -> List[str]:
    """Returns the Orange input or output names, not the Ewoks names."""
    signals = get_signal_list(orange_widget, direction)
    return [signal.name for signal in signals]


def signal_ewoks_to_orange_name(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
    ewoksname: str,
) -> str:
    """Returns the Orange input or output name."""
    signal_container = _get_signal_container(orange_widget, direction)
    signal_dict = _get_signal_ewoks_dict(signal_container)
    if ewoksname not in signal_dict:
        raise RuntimeError(
            f"{ewoksname} is not a signal of {signal_container} of {orange_widget}"
        )
    return signal_dict[ewoksname].name


def signal_orange_to_ewoks_name(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
    orangename: str,
) -> str:
    """Returns the Ewoks name or the `Inputs` or `Outputs` attribute name."""
    signal_container = _get_signal_container(orange_widget, direction)
    signal_dict = _get_signal_ewoks_dict(signal_container)
    for ewoks_or_attr_name, signal in signal_dict.items():
        if signal.name == orangename:
            return ewoks_or_attr_name
    raise RuntimeError(f"{orangename} is not a signal of {signal_container}")


def get_signal(
    orange_widget: Union[OWBaseWidget, Type[OWBaseWidget]],
    direction: Literal["inputs", "outputs"],
    ewoksname: str,
) -> Union[Input, Output]:
    signal_container = _get_signal_container(orange_widget, direction)
    signal_dict = _get_signal_ewoks_dict(signal_container)
    if ewoksname not in signal_dict:
        raise ValueError(
            f"{orange_widget.__name__} does not have {ewoksname!r} in the {direction.title()!r}"
        )
    return signal_dict[ewoksname]


def _receive_dynamic_input(name: str) -> Callable:
    setter_name = f"{name}_ewoks_input_setter"

    def setter(self, value):
        # Called by the SignalManager as a result of calling
        # `send` on an upstream output.
        self.set_dynamic_input(name, value)

    setter.__name__ = setter_name
    return setter


def validate_signals(
    namespace: dict, direction: Literal["inputs", "outputs"], name_to_ignore=tuple()
) -> None:
    """Namespace of an Ewoks-Orange widget class (i.e. not a native Orange widget):

    - Ensure that for each Ewoks Task input and output there is an Orange widget signal.
    - Ensure that the `Inputs`/`Outputs` namespace key exist.
    - Oasys: ensure that the `inputs`/`outputs` namespace key exist.
    """
    ewoks_task = namespace["ewokstaskclass"]
    if direction == "inputs":
        signal_class = Input
        ewoks_names = ewoks_task.input_names()
        ewoks_model = ewoks_task.input_model()
        is_input = True
    elif direction == "outputs":
        signal_class = Output
        ewoks_names = ewoks_task.output_names()
        ewoks_model = ewoks_task.output_model()
        is_input = False
    else:
        raise ValueError(f"{direction=}")
    ewoks_names = tuple(name for name in ewoks_names if name not in name_to_ignore)

    # Do not allow old-style (list instead of signal class) signal definition in Ewoks-Orange widgets.
    signal_container_name = direction.title()
    if direction in namespace:
        raise ValueError(
            f"Use an {signal_container_name!r} class instead of an {direction!r} list"
        )

    # Ensure signal container attribute exists
    if signal_container_name not in namespace:
        namespace[signal_container_name] = type(signal_container_name, (), {})

    # Orange signals
    signal_container_class = namespace[signal_container_name]
    signals_dict = _get_signal_ewoks_dict(signal_container_class)

    # Concatenate Ewoks task-only IO and Orange signals to get the complete list
    ewoks_names = tuple(list(ewoks_names) + list(signals_dict.keys()))

    # Validate signal container
    signals_attrs = list()
    new_signals_class = False
    for ewoksname in ewoks_names:
        # Orange signal for the Ewoks variable.
        signal = signals_dict.get(ewoksname, None)
        if signal is None:
            data_type = _pydantic_model_field_type(ewoks_model, ewoksname)
            doc = _pydantic_model_field_doc(
                ewoks_model,
                ewoksname,
            )
            orangename = ewoksname
            signal = signal_class(name=orangename, type=data_type, doc=doc)
            new_signals_class = True

        if is_input:
            # Create a handler for the input value provided
            # by upstream nodes at runtime, unless already provided.
            handler: str = signal.handler
            if not handler or handler not in namespace:
                setter = _receive_dynamic_input(ewoksname)
                handler = setter.__name__
                namespace[handler] = signal(setter)  # does input.handler = handler

        # Ensure the signal knows about the Ewoks parameter name as well
        signal.ewoksname = ewoksname

        signals_attrs.append((ewoksname, signal))

    # Ensure Ewoks order
    for i, (_, signal) in enumerate(signals_attrs, 1):
        signal._seq_id = i

    # Replace signal class when needed
    if new_signals_class:
        signal_container_class = type(signal_container_name, (), dict(signals_attrs))
        namespace[signal_container_name] = signal_container_class

    # Oasys needs old-style signal definitions
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        if len(namespace.get(direction, [])) != len(ewoks_names):
            namespace[direction] = _oldstyle_signal_list(signal_container_class)


def _pydantic_model_field_type(
    model: Optional[Type[BaseModel]], field_name: str, default_data_type=object
) -> type:
    if model is None:
        return default_data_type
    field_info = model.model_fields.get(field_name, None)
    if field_info is None:
        return default_data_type
    origin = get_origin(field_info.annotation)
    valid_union_types = [
        Union,
    ]
    if has_UnionType:
        valid_union_types.append(UnionType)
    if origin is None:
        # if unsupported ()
        return field_info.annotation
    elif origin in (list, tuple):
        return origin
    elif ORANGE_VERSION != ORANGE_VERSION.oasys_fork and origin in valid_union_types:
        # Handle Union types (including Optional)
        # This feature is only accessible for "recent" orange version and not for OASYS where type must be a scalar
        args = get_args(field_info.annotation)
        non_none_args = [arg for arg in args if arg is not type(None)]
        return tuple(non_none_args)
    else:
        # other cases
        return object


def _pydantic_model_field_doc(
    model: Optional[Type[BaseModel]], field_name: str, default_doc=None
) -> Optional[str]:
    if model is None:
        return default_doc
    field_info = model.model_fields.get(field_name, None)
    if field_info is None:
        return default_doc
    try:
        return field_info.description
    except AttributeError:
        return default_doc


def _oldstyle_signal_container(
    signals: List[Tuple[str]], direction: Literal["inputs", "outputs"]
) -> type:
    """Convert

    .. code-block:: python

        inputs = [("A", object, ""), ("B", object, "")]  # list of tuples or dicts

    to

    .. code-block:: python

        class Inputs:
            a = Input("A", object)
            b = Input("B", object)
    """
    if direction == "inputs":
        prefix = "input"
        signal_class = Input
    elif direction == "outputs":
        prefix = "output"
        signal_class = Output
    else:
        raise ValueError(f"{direction=}")
    names = [f"{prefix}{i}" for i in range(len(signals))]
    values = [_oldstyle_instantiate_signal(signal_class, signal) for signal in signals]
    attrs = dict(zip(names, values))
    return type(direction.title(), (), attrs)


def _oldstyle_signal_list(input_container_class) -> List[str]:
    """Convert

    .. code-block:: python

        class Inputs:
            a = Input("A", object)
            b = Input("B", object)

    to

    .. code-block:: python

        inputs = [("A", object, "")]
    """
    signals = _get_signal_list_from_container(input_container_class)
    return [signal.as_tuple() for signal in signals]


def _oldstyle_instantiate_signal(
    signal_class: Union[Type[Input], Type[Output]], data: Union[tuple, dict]
) -> Union[Input, Output]:
    if isinstance(data, tuple):
        signal = signal_class(*data, ewoksname=data[0])
    elif isinstance(data, dict):
        signal = signal_class(**data, ewoksname=data["name"])
    elif isinstance(data, _InputSignal):
        names = "name", "type", "id", "doc", "replaces"
        data_dict = {name: getattr(data, name) for name in names if hasattr(data, name)}
        signal = signal_class(**data_dict)
        if hasattr(data, "handler"):
            signal.handler = getattr(data, "handler")
        signal.ewoksname = signal.name
    elif isinstance(data, _OutputSignal):
        names = "name", "type", "id", "doc", "replaces"
        data_dict = {name: getattr(data, name) for name in names if hasattr(data, name)}
        signal = signal_class(**data_dict)
        signal.ewoksname = signal.name
    else:
        raise TypeError(type(data))
    return signal



# flake8-in-file-ignores:
import warnings

from ..gui.owwidgets.base import OWBaseWidget  # noqa: F401
from ..gui.owwidgets.base import OWEwoksBaseWidget  # noqa: F401
from ..gui.owwidgets.base import OWWidget  # noqa: F401
from ..gui.owwidgets.meta import ow_build_opts
from ..gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ..gui.owwidgets.threaded import OWEwoksWidgetOneThread
from ..gui.owwidgets.threaded import OWEwoksWidgetOneThreadPerRun
from ..gui.owwidgets.threaded import OWEwoksWidgetWithTaskStack
from ..gui.owwidgets.types import is_ewoks_widget  # noqa: F401
from ..gui.owwidgets.types import is_ewoks_widget_class  # noqa: F401
from ..gui.owwidgets.types import is_native_widget  # noqa: F401
from ..gui.owwidgets.types import is_native_widget_class  # noqa: F401
from ..gui.owwidgets.types import is_orange_widget  # noqa: F401
from ..gui.owwidgets.types import is_orange_widget_class  # noqa: F401
from ..gui.utils import invalid_data  # noqa: F401

__all__ = [
    "OWEwoksWidgetNoThread",
    "OWEwoksWidgetOneThread",
    "OWEwoksWidgetOneThreadPerRun",
    "OWEwoksWidgetWithTaskStack",
    "ow_build_opts",
]

warnings.warn(
    f"The '{__name__}' module is deprecated and will be removed in a future release. "
    "Please migrate to the new 'ewoksorange.gui.owwidgets.*' modules.",
    DeprecationWarning,
    stacklevel=2,
)



"""
Summarizers for Orange signal display integration.
"""

from ...orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    summarize = None
    PartialSummary = None
else:
    from orangewidget.utils.signals import PartialSummary
    from orangewidget.utils.signals import summarize

from ewokscore.variable import Variable

if summarize is not None:

    @summarize.register(Variable)
    def summarize_variable(var: Variable):
        """
        Provide a short summary for Ewoks Variable instances for Orange UI.

        :param var: The Variable to summarize.
        :return: PartialSummary describing the variable.
        """
        if var.is_missing():
            dtype = var.value
        else:
            dtype = type(var.value).__name__
        desc = f"ewoks variable ({dtype})"
        return PartialSummary(desc, desc)

    @summarize.register(object)
    def summarize_object(value: object):
        """
        Provide a default summary for arbitrary objects.

        :param value: The object to summarize.
        :return: PartialSummary describing the object's type.
        """
        return PartialSummary(str(type(value)), str(type(value)))



import numpy
from ewokscore.task import Task
from ewokscore.tests.examples.tasks.sumlist import SumList
from ewokscore.tests.examples.tasks.sumtask import SumTask


class SumTaskTest(SumTask):
    pass


class PrintSum(Task, input_names=["sum"]):
    def run(self):
        if self.inputs.sum is None:
            raise ValueError("'value' should be provided")
        print("input value is", self.inputs.sum)


class SumList1(SumList):
    # as each OW request his own Task we need to create this "dummy class"
    pass


class SumList2(SumList):
    # as each OW request his own Task we need to create this "dummy class"
    pass


class SumList3(SumList):
    # as each OW request his own Task we need to create this "dummy class"
    pass


class SumList4(SumList):
    # as each OW request his own Task we need to create this "dummy class"
    pass


class GenerateList(Task, input_names=["length"], output_names=["list"]):
    def run(self):
        if self.inputs.length is None:
            raise ValueError("length should be provided")
        self.outputs.list = numpy.random.random(self.inputs.length) * 100.0



import os
from contextlib import contextmanager
from typing import Any
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from ewokscore.graph.serialize import GraphRepresentation

from .owscheme import ewoks_to_ows
from .owscheme import ows_to_ewoks


def get_representation(
    graph: Any, representation: Optional[Union[GraphRepresentation, str]] = None
) -> Optional[str]:
    if (
        representation is None
        and isinstance(graph, str)
        and graph.lower().endswith(".ows")
    ):
        representation = "ows"
    return representation


@contextmanager
def ows_file_context(
    graph,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    task_options: Optional[dict] = None,
    error_on_duplicates: bool = True,
    tmpdir: Optional[str] = None,
) -> Generator[str, None, None]:
    """Yields an .ows file path (temporary file when not alread an .ows file)"""
    if load_options is None:
        load_options = dict()
    representation = get_representation(
        graph, representation=load_options.get("representation")
    )
    if representation == "ows":
        ows_filename = graph
        if inputs or varinfo or execinfo or task_options:
            # Already an .ows file but we need to inject data so that
            # `OWEwoksBaseWidget` can retrieve it in `_get_task_arguments`
            # to instantiate an Ewoks tasks.
            # See `OwsNodeWrapper` on how this information gets passed.
            graph = ows_to_ewoks(ows_filename, **load_options)
            basename = os.path.splitext(os.path.basename(ows_filename))[0]
            if tmpdir:
                tmp_filename = os.path.abspath(
                    os.path.join(str(tmpdir), f"{basename}_mod.ows")
                )
            else:
                tmp_filename = os.path.abspath(f"{basename}_mod.ows")
            try:
                ewoks_to_ows(
                    graph,
                    tmp_filename,
                    inputs=inputs,
                    varinfo=varinfo,
                    execinfo=execinfo,
                    task_options=task_options,
                    error_on_duplicates=error_on_duplicates,
                )
                yield tmp_filename
            finally:
                if os.path.exists(tmp_filename):
                    os.remove(tmp_filename)
        else:
            # Already an .ows file
            yield ows_filename
    else:
        # Convert to an .ows file before launching the GUI
        if tmpdir:
            tmp_filename = os.path.abspath(
                os.path.join(str(tmpdir), "ewoks_workflow_tmp.ows")
            )
        else:
            tmp_filename = os.path.abspath("ewoks_workflow_tmp.ows")
        try:
            ewoks_to_ows(
                graph,
                tmp_filename,
                inputs=inputs,
                varinfo=varinfo,
                execinfo=execinfo,
                task_options=task_options,
                error_on_duplicates=error_on_duplicates,
                **load_options,
            )
            yield tmp_filename
        finally:
            if os.path.exists(tmp_filename):
                os.remove(tmp_filename)



import pytest
from ewokscore.task import Task
from ewoksutils.import_utils import qualname

from ..gui.orange_utils.signals import Input
from ..gui.owwidgets.base import OWWidget
from ..gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ..gui.owwidgets.registration import register_owwidget
from ..gui.workflows.owscheme import ewoks_to_ows
from ..orange_version import ORANGE_VERSION


class Mother(int): ...


class SubClass(Mother): ...


if ORANGE_VERSION != ORANGE_VERSION.oasys_fork:
    # else with oasys we need to provide the 'handler' mechanism
    class NativeWidget(OWWidget):
        name = "native widget"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._data = None

        class Inputs:
            data = Input("data", type=Mother)

        @Inputs.data
        def data_received(self, data):
            self._data = data


class EwoksTask(
    Task,
    input_names=(),
    output_names=("data",),
):
    def run(self):
        self.outputs.data = SubClass(2)


class EwoksOrangeWidget(OWEwoksWidgetNoThread, ewokstaskclass=EwoksTask):
    name = "ewoks widget"


@pytest.mark.skipif(
    ORANGE_VERSION == ORANGE_VERSION.oasys_fork, reason="hanging with oasys binding."
)
def test_dynamic_link(tmpdir, ewoks_orange_canvas):
    """Test that a dynamic link in orange will be processed as expected."""
    # Create an Orange workflows
    workflow = {
        "graph": {
            "id": "ewoksgraph",
            "label": "Ewoks workflow 'ewoksgraph'",
            "schema_version": "1.1",
        },
        "links": [
            {
                "data_mapping": [{"source_output": "data", "target_input": "data"}],
                "source": "0",
                "target": "1",
            }
        ],
        "nodes": [
            {
                "id": "0",
                "task_identifier": qualname(EwoksTask),
                "task_type": "class",
            },
            {
                "id": "1",
                "task_generator": "ewoksorange.bindings.taskwrapper.owwidget_task_wrapper",
                "task_identifier": qualname(NativeWidget),
                "task_type": "generated",
            },
        ],
    }
    destination = str(tmpdir / "ewoksgraph.ows")
    ewoks_to_ows(workflow, destination)

    for widget in (NativeWidget, EwoksOrangeWidget):
        register_owwidget(
            widget_class=widget,
            package_name="ewoksorange",
            category_name="test",
            project_name="ewoksorange",
        )

    # Load and execute the orange workflow
    ewoks_orange_canvas.load_ows(destination)
    ewoks_orange_canvas.start_workflow()

    ewoks_orange_canvas.wait_widgets(timeout=10)
    native_widget = next(ewoks_orange_canvas.widgets_from_name("1"))
    assert native_widget._data == 2



from typing import Tuple

from ewokscore import Task
from ewokscore import TaskWithProgress
from ewoksutils.import_utils import qualname

from ewoksorange.gui.owwidgets import registration
from ewoksorange.gui.owwidgets.base import OWEwoksBaseWidget
from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetOneThread

from . import widgets

NAME = "No widgets"

DESCRIPTION = "Ewoks tasks without widgets"

LONG_DESCRIPTION = "Widgets for Ewoks tasks that come with a bare Ewoks installation"

ICON = "icons/category.png"

BACKGROUND = "light-blue"

_PROJECT_NAME = "ewoksorange"  # to avoid the "missing addon" error message when opening an ewoks workflow with non-existing widgets

_DEFAULT_WIDGET_CLASSES = dict()


def register_owwidget(widget_class, discovery_object=None):
    package_name = __name__
    category_name = "Ewoks Without Widgets"
    project_name = _PROJECT_NAME

    registration.register_owwidget(
        widget_class,
        package_name,
        category_name,
        project_name,
        discovery_object=discovery_object,
    )


def default_owwidget_class(task_class: Task) -> Tuple[OWEwoksBaseWidget, str]:
    widget_class = _DEFAULT_WIDGET_CLASSES.get(task_class, None)
    if widget_class is not None:
        return widget_class, _PROJECT_NAME

    # Create the widget class
    if issubclass(TaskWithProgress, task_class):
        basecls = OWEwoksWidgetOneThread
    else:
        basecls = OWEwoksWidgetNoThread

    class DefaultOwWidget(basecls, ewokstaskclass=task_class):
        name = qualname(
            task_class
        )  # Allows recreating `DefaultOwWidget` when loading .ows file
        description = f"Orange widget is missing for Ewoks task {task_class.__name__}"
        icon = "icons/nowidget.svg"
        want_main_area = False

        def __init__(self, *args, **kw):
            super().__init__(*args, **kw)
            self._init_control_area()

    widget_class = DefaultOwWidget

    # Add the class to the 'widgets' module
    widget_class.__name__ += "_" + task_class.class_registry_name().replace(".", "_")
    widget_class.__module__ = widgets.__name__
    setattr(widgets, widget_class.__name__, widget_class)

    # Register the widget class
    _DEFAULT_WIDGET_CLASSES[task_class] = widget_class
    register_owwidget(widget_class)
    return widget_class, _PROJECT_NAME


def widget_discovery(discovery):
    for widget_class in _DEFAULT_WIDGET_CLASSES.values():
        register_owwidget(widget_class, discovery_object=discovery)


def global_cleanup_ewoksnowidget():
    """Remove all widget declarations."""
    # Note: the discovery object owns a widget registry which keeps references
    # to all discovered widget classes. The widgets from `ewoksnowidget` are
    # never registered with a global widget registry (so not visible in the
    # widget panel in the canvas) so there is no need to remove the widgets
    # from the registry used during discovery.
    for widget_class in _DEFAULT_WIDGET_CLASSES.values():
        delattr(widgets, widget_class.__name__)
    _DEFAULT_WIDGET_CLASSES.clear()



from typing import Optional

from AnyQt.QtCore import QThread

from .base import TaskExecutor


class ThreadedTaskExecutor(QThread, TaskExecutor):
    """Create and execute an Ewoks task in a dedicated thread."""

    def run(self) -> None:
        self.execute_task()

    def stop(self, timeout: Optional[float] = None, wait: bool = False) -> None:
        """Stop the current thread"""
        self.blockSignals(True)
        if wait:
            if timeout:
                self.wait(timeout * 1000)
            else:
                self.wait()
        if self.isRunning():
            self.quit()

    def cancel_running_task(self):
        """
        cancel current processing.
        The targetted EwoksTask must have implemented the 'cancel' function
        """
        if self.current_task is not None:
            self.current_task.cancel()



import pytest
from ewokscore import execute_graph

from ..gui.workflows.owscheme import ows_to_ewoks
from ..orange_version import ORANGE_VERSION

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


def test_sumtask_tutorial_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumtask_tutorial.ows")
    assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_sumtask_tutorial_without_qt(qtapp):
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumtask_tutorial.ows")
    assert_sumtask_tutorial_without_qt(filename)


def test_list_operations_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumlist_tutorial.ows")
    assert_sumlist_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_list_operations_without_qt(qtapp):
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumlist_tutorial.ows")
    assert_sumlist_tutorial_without_qt(filename)


def test_mixed_tutorial_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewokstest import tutorials

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        workflow = "mixed_tutorial.ows"
    elif ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        workflow = "mixed_tutorial_oasys.ows"
    else:
        pytest.skip("Requires the Orange3 or Oasys1 python script widget")

    filename = resource_files(tutorials).joinpath(workflow)
    assert_mixed_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_mixed_tutorial_without_qt(qtapp):
    from orangecontrib.ewokstest import tutorials

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        workflow = "mixed_tutorial.ows"
    elif ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        workflow = "mixed_tutorial_oasys.ows"
    else:
        pytest.skip("Requires the Orange3 or Oasys1 python script widget")

    filename = resource_files(tutorials).joinpath(workflow)
    assert_mixed_tutorial_without_qt(filename)


def assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    widgets = list(ewoks_orange_canvas.widgets_from_name("task6"))
    results = widgets[0].get_task_output_values()
    assert results == {"result": 16}

    ewoks_orange_canvas.load_ows(str(filename))
    ewoks_orange_canvas.set_input_values(
        [{"label": "task1", "name": "b", "value": "wrongtype"}]
    )
    ewoks_orange_canvas.start_workflow()
    with pytest.raises(TypeError):
        # Note: we get the original error, not "RuntimeError: Task 'task1' failed"
        ewoks_orange_canvas.wait_widgets(timeout=10)


def assert_sumtask_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    results = execute_graph(graph, output_tasks=True)
    assert results["5"].get_output_values() == {"result": 16}


def assert_sumlist_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))

    # Remove artificial delay for this test
    for widget in ewoks_orange_canvas.iter_widgets():
        if "delay" in widget.get_default_input_names():
            widget.update_default_inputs(delay=0)

    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)

    wgenerator = list(ewoks_orange_canvas.widgets_from_name("List generator"))[0]
    results = wgenerator.get_task_output_values()
    listsum = sum(results["list"])

    widgets = list(ewoks_orange_canvas.widgets_from_name("Print list sum"))
    widgets += list(ewoks_orange_canvas.widgets_from_name("Print list sum (1)"))
    widgets += list(ewoks_orange_canvas.widgets_from_name("Print list sum (2)"))
    for w in widgets:
        results = {name: var.value for name, var in w.get_task_inputs().items()}
        assert results == {"sum": listsum}


def assert_sumlist_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)

    # Remove artificial delay for this test
    for attrs in graph.graph.nodes.values():
        for adict in attrs.get("default_inputs", list()):
            if adict["name"] == "delay":
                adict["value"] = 0

    results = execute_graph(graph, output_tasks=True)
    listsum = sum(results["0"].get_output_values()["list"])
    for i in [4, 5, 6]:
        assert results[str(i)].get_input_values() == {"sum": listsum}


def assert_mixed_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    widget = ewoks_orange_canvas.widget_from_id("2")
    results = widget.get_task_output_values()
    assert results == {"result": 3}


def assert_mixed_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    tasks = execute_graph(graph, output_tasks=True)
    results = tasks["2"].get_output_values()
    assert results == {"result": 3}



import importlib
import sys
import warnings
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Union

import ewokscore
from ewokscore.graph import TaskGraph
from ewokscore.graph.serialize import GraphRepresentation

from ..gui.canvas.main import main as launchcanvas
from ..gui.workflows import owscheme
from ..gui.workflows.representation import get_representation
from ..gui.workflows.representation import ows_file_context


@ewokscore.execute_graph_decorator(engine="orange")
def execute_graph(
    graph: Any,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    task_options: Optional[dict] = None,
    outputs: Optional[List[dict]] = None,
    merge_outputs: Optional[bool] = True,
    error_on_duplicates: bool = True,
    tmpdir: Optional[str] = None,
) -> None:
    if outputs:
        raise ValueError("The Orange3 binding cannot return any results")
    with ows_file_context(
        graph,
        inputs=inputs,
        load_options=load_options,
        varinfo=varinfo,
        execinfo=execinfo,
        task_options=task_options,
        error_on_duplicates=error_on_duplicates,
        tmpdir=tmpdir,
    ) as ows_filename:
        argv = [sys.argv[0], ows_filename]
        launchcanvas(argv=argv)


def load_graph(
    graph: Any,
    inputs: Optional[List[dict]] = None,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
    preserve_ows_info: Optional[bool] = True,
    title_as_node_id: Optional[bool] = False,
) -> TaskGraph:
    representation = get_representation(graph, representation=representation)
    if representation == "ows":
        return owscheme.ows_to_ewoks(
            graph,
            inputs=inputs,
            root_dir=root_dir,
            root_module=root_module,
            preserve_ows_info=preserve_ows_info,
            title_as_node_id=title_as_node_id,
        )
    else:
        return ewokscore.load_graph(
            graph,
            inputs=inputs,
            representation=representation,
            root_dir=root_dir,
            root_module=root_module,
        )


def save_graph(
    graph: TaskGraph,
    destination,
    representation: Optional[Union[GraphRepresentation, str]] = None,
    **save_options,
) -> Union[str, dict]:
    representation = get_representation(destination, representation=representation)
    if representation == "ows":
        owscheme.ewoks_to_ows(graph, destination, **save_options)
        return destination
    else:
        return graph.dump(destination, representation=representation, **save_options)


def convert_graph(
    source,
    destination,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    save_options: Optional[dict] = None,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    if save_options is None:
        save_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    return save_graph(graph, destination, **save_options)


__deprecated_submodules__ = {
    "owsconvert": "ewoksorange.bindings.owsconvert",
    "owwidgets": "ewoksorange.bindings.owwidgets",
    "progress": "ewoksorange.bindings.progress",
    "taskwrapper": "ewoksorange.bindings.taskwrapper",
}


def __getattr__(name):
    for full_module in __deprecated_submodules__.values():
        try:
            submod = importlib.import_module(full_module)
            if hasattr(submod, name):
                warnings.warn(
                    f"Accessing '{name}' from '{__name__}' is deprecated. "
                    f"Please import from '{full_module}' instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return getattr(submod, name)
        except ImportError:
            continue

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from .base import OWBaseWidget
from .base import OWEwoksBaseWidget


def is_orange_widget_class(widget_class) -> bool:
    """
    Return True if `widget_class` is an native Orange or Ewoks-Orange widget class.
    """
    return isinstance(widget_class, type) and issubclass(widget_class, OWBaseWidget)


def is_ewoks_widget_class(widget_class) -> bool:
    """
    Return True if `widget_class` is an Ewoks-Orange widget class.
    """
    return isinstance(widget_class, type) and issubclass(
        widget_class, OWEwoksBaseWidget
    )


def is_native_widget_class(widget_class) -> bool:
    """
    Return True if `widget_class` is a native Orange widget class.
    """
    return is_orange_widget_class(widget_class) and not is_ewoks_widget_class(
        widget_class
    )


def is_orange_widget(widget) -> bool:
    """
    Return True if `widget` is an Orange or Ewoks-Orange widget.
    """
    return isinstance(widget, OWBaseWidget)


def is_ewoks_widget(widget) -> bool:
    """
    Return True if `widget` is an Ewoks-Orange widget.
    """
    return isinstance(widget, OWEwoksBaseWidget)


def is_native_widget(widget) -> bool:
    """
    Return True if `widget` is a native Orange widget.
    """
    return is_orange_widget(widget) and not is_ewoks_widget(widget)



import importlib.metadata
from typing import Optional

from packaging.version import Version

try:
    _USE_IMPORTLIB = Version(
        importlib.metadata.version("orange-canvas-core")
    ) >= Version("0.2.0")
    # orange-canvas-core>=0.2.0
    # orange-widget-base>=4.23.0
    # orange3>=3.37.0
    #
    # https://github.com/biolab/orange-canvas-core/pull/289
    # https://github.com/biolab/orange-widget-base/pull/260
    # https://github.com/biolab/orange3/pull/6655
except importlib.metadata.PackageNotFoundError as ex:
    try:
        _USE_IMPORTLIB = Version(
            importlib.metadata.version("oasys-canvas-core")
        ) >= Version("1.0.10")
        # Oasys-Canvas-Core   1.0.10
        # Oasys-Widget-Core   1.0.5
        # OASYS1              1.2.148
    except importlib.metadata.PackageNotFoundError:
        raise ex


if _USE_IMPORTLIB:
    from ewokscore.entry_points import EntryPoint  # noqa F401
    from ewokscore.entry_points import entry_points  # noqa F401

    def get_distribution(
        name: str, raise_error: bool = False
    ) -> Optional[importlib.metadata.Distribution]:
        try:
            return importlib.metadata.Distribution.from_name(name)
        except importlib.metadata.PackageNotFoundError:
            if raise_error:
                raise

    def get_distribution_name(distribution: importlib.metadata.Distribution) -> str:
        return distribution.name

else:
    import pkg_resources
    from pkg_resources import EntryPoint  # noqa F401
    from pkg_resources import iter_entry_points as entry_points  # noqa F401

    def get_distribution(
        name: str, raise_error: bool = False
    ) -> Optional[pkg_resources.Distribution]:
        try:
            return pkg_resources.get_distribution(name)
        except pkg_resources.DistributionNotFound:
            if raise_error:
                raise

    def get_distribution_name(distribution: pkg_resources.Distribution) -> str:
        return distribution.project_name



from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetWithTaskStack
from ewoksorange.tests.examples.tasks import SumList3
from ewoksorange.tests.examples.widgets import WidgetMixin


class SumListWithTaskStack(
    WidgetMixin, OWEwoksWidgetWithTaskStack, ewokstaskclass=SumList3
):
    """
    Simple demo class that will process task with a FIFO stack and one thread
    connected with the stack
    """

    name = "SumList with one thread and a stack"
    description = "Sum all elements of a list using a thread and a stack"
    icon = "icons/mywidget.svg"
    want_main_area = True



import functools
import gc
import logging
import warnings

import pytest

from orangecontrib.ewoksnowidget import global_cleanup_ewoksnowidget
from orangecontrib.ewokstest import enable_ewokstest_category

from ..gui.canvas.handler import OrangeCanvasHandler
from ..gui.qt_utils.app import get_all_qtwidgets
from ..gui.qt_utils.app import qtapp_context
from ..orange_version import ORANGE_VERSION

logger = logging.getLogger(__name__)


def global_cleanup_orange():
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        pass
    else:
        from orangecanvas.document.suggestions import Suggestions

        Suggestions.instance = None


def global_cleanup_pytest():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for obj in gc.get_objects():
            if isinstance(obj, logging.LogRecord):
                obj.exc_info = None  # traceback keeps frames which keep locals


def collect_garbage(app):
    app.processEvents()
    while gc.collect():
        app.processEvents()


def safe_session_fixture(fixture):
    """Use instead of `pytest.fixture` to ensure the session fixture is executed only once."""
    return_value = None

    @functools.wraps(fixture)
    def wrapper(*args, **kw):
        nonlocal return_value

        if return_value is None:
            gen = fixture(*args, **kw)
            return_value = next(gen)
            try:
                yield return_value
            finally:
                try:
                    next(gen)
                except StopIteration:
                    pass
        else:
            yield return_value

    return pytest.fixture(scope="session")(wrapper)


@safe_session_fixture
def qtapp():
    enable_ewokstest_category()
    with qtapp_context() as app:
        assert app is not None
        yield app
    collect_garbage(app)
    global_cleanup_ewoksnowidget()
    global_cleanup_orange()
    global_cleanup_pytest()
    collect_garbage(app)
    warn_qtwidgets_alive()


@pytest.fixture(scope="session")
def raw_ewoks_orange_canvas(qtapp):
    with OrangeCanvasHandler() as handler:
        yield handler


@pytest.fixture()
def ewoks_orange_canvas(raw_ewoks_orange_canvas):
    yield raw_ewoks_orange_canvas
    try:
        raw_ewoks_orange_canvas.scheme.ewoks_finalize()
    except AttributeError:
        pass


def warn_qtwidgets_alive():
    widgets = get_all_qtwidgets()
    if widgets:
        logger.warning("%d remaining widgets after tests", len(widgets))



"""
Synchronous (no-thread) Ewoks widget implementation.
"""

import logging
from typing import Optional

from ..concurrency.base import TaskExecutor
from .base import OWEwoksBaseWidget
from .meta import ow_build_opts

_logger = logging.getLogger(__name__)


class OWEwoksWidgetNoThread(OWEwoksBaseWidget, **ow_build_opts):
    """
    Widget that creates and executes an Ewoks Task synchronously on the main thread.

    Use this for lightweight tasks that won't block the UI.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the no-thread widget and preparer a TaskExecutor.
        """
        super().__init__(*args, **kwargs)
        self.__task_executor = TaskExecutor(self.ewokstaskclass)

    def _execute_ewoks_task(self, propagate: bool, log_missing_inputs: bool) -> None:
        """
        Create and execute the Task synchronously.

        :param propagate: Whether to propagate outputs after execution.
        :param log_missing_inputs: Whether to log missing input warnings.
        """
        self.__task_executor.create_task(
            log_missing_inputs=log_missing_inputs, **self._get_task_arguments()
        )
        try:
            self.__task_executor.execute_task()
        except Exception as e:
            _logger.error(f"task failed: {e}", exc_info=True)
        try:
            self.__post_task_exception = None
            if propagate:
                self.propagate_downstream()
        finally:
            self._output_changed()

    @property
    def task_succeeded(self) -> Optional[bool]:
        """Return True if last task succeeded, False if failed, None if never run."""
        return self.__task_executor.succeeded

    @property
    def task_done(self) -> Optional[bool]:
        """Return True if last task finished (success/failure), None if never run."""
        return self.__task_executor.done

    @property
    def task_exception(self) -> Optional[Exception]:
        """Return the exception raised during last task execution, if any."""
        return self.__task_executor.exception

    def get_task_outputs(self) -> dict:
        """Return output variables produced by the last executed task."""
        return self.__task_executor.output_variables



from numbers import Number
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Type

from ewokscore import Task
from ewokscore.variable import value_from_transfer
from ewoksutils.import_utils import import_qualname
from ewoksutils.import_utils import qualname

from ..orange_utils import _signals
from ..orange_utils import settings
from ..orange_utils.signal_manager import SignalManagerWithoutScheme
from ..orange_utils.signal_manager import set_input_value
from ..owwidgets.base import OWBaseWidget
from ..owwidgets.base import OWEwoksBaseWidget
from ..owwidgets.types import is_ewoks_widget_class
from ..owwidgets.types import is_native_widget_class
from ..qt_utils.app import QtEvent
from ..qt_utils.app import ensure_qtapp
from ..qt_utils.app import process_qtapp_events
from ..utils import invalid_data


def owwidget_task_wrapper(widget_qualname: str) -> Task:
    """Create a task that does the computation through an orange widget.
    When the widget is an ewoks widget, still use the widget and not
    the corresponding task class directly.
    """
    widget_class = import_qualname(widget_qualname)
    registry_name = widget_qualname + ".wrapper"
    if registry_name in Task.get_subclass_names():
        return Task.get_subclass(registry_name)

    if is_ewoks_widget_class(widget_class):
        return _ewoks_owwidget_task_wrapper(registry_name, widget_class)
    elif is_native_widget_class(widget_class):
        return _native_owwidget_task_wrapper(registry_name, widget_class)
    else:
        raise TypeError(widget_class, "expected to be an OWWidget")


OWWIDGET_TASKS_GENERATOR = qualname(owwidget_task_wrapper)


def _ewoks_owwidget_task_wrapper(
    registry_name: str, widget_class: OWEwoksBaseWidget
) -> Task:
    """Wrap an Ewoks widget with an Ewoks task"""
    all_input_names = widget_class.get_input_names()
    try:
        ewokstaskclass = widget_class.ewokstaskclass
        input_names = ewokstaskclass.required_input_names()
        optional_input_names = ewokstaskclass.optional_input_names()
        expected = set(input_names) | set(optional_input_names)
        assert all_input_names == expected
    except AttributeError:
        input_names = all_input_names
        optional_input_names = tuple()
    output_names = widget_class.get_output_names()

    class WrapperTask(
        Task,
        input_names=input_names,
        optional_input_names=optional_input_names,
        output_names=output_names,
        registry_name=registry_name,
    ):
        def run(self):
            output_values = execute_ewoks_owwidget(
                widget_class, inputs=self.get_input_values()
            )
            for k, v in output_values.items():
                self.output_variables[k].value = v

    return WrapperTask


def _native_owwidget_task_wrapper(
    registry_name: str, widget_class: OWBaseWidget
) -> Task:
    """Wrap a native Orange widget with an Ewoks task"""
    input_signals = _signals.get_signal_list(widget_class, "inputs")
    optional_input_names = {input.ewoksname for input in input_signals}
    output_signals = _signals.get_signal_list(widget_class, "outputs")
    output_names = {output.ewoksname for output in output_signals}
    input_names = tuple()

    class WrapperTask(
        Task,
        input_names=input_names,
        optional_input_names=optional_input_names,
        output_names=output_names,
        registry_name=registry_name,
    ):
        def run(self):
            output_values = execute_native_owwidget(
                widget_class, inputs=self.get_input_values()
            )
            for k, v in output_values.items():
                self.output_variables[k].value = v

    return WrapperTask


def instantiate_owwidget(
    widget_class: Type[OWBaseWidget],
    signal_manager=None,
    stored_settings: Optional[Mapping] = None,
    **widget_init_params,
):
    if stored_settings:
        stored_settings = {
            k: v
            for k, v in stored_settings.items()
            if not invalid_data.is_invalid_data(v)
        }
    widget = widget_class.__new__(
        widget_class, signal_manager=signal_manager, stored_settings=stored_settings
    )
    widget.__init__(**widget_init_params)
    return widget


def execute_ewoks_owwidget(
    widget_class: Type[OWEwoksBaseWidget],
    inputs: Optional[Mapping] = None,
    timeout: Optional[Number] = None,
    **widget_init_params,
) -> dict:
    """This is the equivalent of the execution of the associated Ewoks task

    .. code-block::python

        task = task_cls(inputs=inputs)
        task.execute()
        return task.get_output_values()

    but instead execute it like Orange would do it (using Qt signals).

    It is used for testing Ewoks Orange widgets.
    """
    ensure_qtapp()
    result = dict()
    exception = None
    widget = instantiate_owwidget(widget_class, **widget_init_params)

    try:
        # Receive and store results
        outputsReceived = QtEvent()

        def _output_cb():
            nonlocal exception

            try:
                exception = widget.task_exception or widget.post_task_exception
                result.update(widget.get_task_output_values())
            finally:
                outputsReceived.set()

        widget.task_output_changed_callbacks.append(_output_cb)

        # Call the input setters
        if inputs:
            signals = _signals.get_signal_list(widget, "inputs")

            for index, signal in enumerate(signals):
                key = signal.ewoksname
                if key in inputs:
                    value = value_from_transfer(inputs[key])
                    set_input_value(widget, signal, value, index)

        # Start calculation
        try:
            widget.handleNewSignals()
        except Exception:
            # Widget executes everything in the current thread
            # The exception should have been captured by _output_cb
            # If not there is an implementation problem and we raise
            if exception is None:
                raise

        # Wait for the result
        if not outputsReceived.wait(timeout=timeout):
            raise TimeoutError(f"{timeout} sec")

        # Raise task exception
        if exception is not None:
            raise exception
    finally:
        widget.close()
    return result


def execute_native_owwidget(
    widget_class: Type[OWBaseWidget],
    inputs: Optional[Mapping] = None,
    timeout: Optional[Number] = None,
    settings: Optional[Mapping] = None,
    **widget_init_params,
):
    """This is the equivalent of `execute_ewoks_owwidget` but then for native Orange widget
    instead of Ewoks Orange Widgets.

    It is used to execute native Orange widgets with another Ewoks execution engine than Orange.
    """
    ensure_qtapp()
    result = dict()

    output_signals = _signals.get_signal_list(widget_class, "outputs")
    ewoks_to_orange_namemap = {
        signal.ewoksname: signal.name for signal in output_signals
    }

    input_list, stored_settings = _parse_input_values(widget_class, inputs)

    # Create widget with the proper settings
    if settings:
        stored_settings.update(settings)
    widget = instantiate_owwidget(
        widget_class,
        signal_manager=SignalManagerWithoutScheme(),
        stored_settings=stored_settings,
        **widget_init_params,
    )

    exception = None
    try:
        # Call input setters
        for index, (signal, value) in enumerate(input_list):
            try:
                set_input_value(widget, signal, value, index)
            except Exception as ex:
                exception = exception or ex

        # Start calculation
        try:
            widget.handleNewSignals()
        except Exception as ex:
            exception = exception or ex

        # Wait for the result
        process_qtapp_events()

        # Raise task exception
        if exception is not None:
            raise exception

        # Fetch outputs
        for ewoks_name, orange_name in ewoks_to_orange_namemap.items():
            value = widget.signalManager.get_output_value(
                widget, orange_name, timeout=timeout
            )
            result[ewoks_name] = value
    finally:
        widget.close()
    return result


def _parse_input_values(
    widget_class, inputs: Optional[Mapping] = None
) -> Tuple[List, dict]:
    used_values = set()
    settings_dict = dict()
    input_list = list()
    if inputs is None:
        inputs = dict()

    # Values corresponding to settings
    setting_names = list(settings.get_settings(widget_class))
    for ewoksname in setting_names:
        if ewoksname not in inputs:
            continue
        used_values.add(ewoksname)
        value = value_from_transfer(inputs[ewoksname])
        settings_dict[ewoksname] = value

    # Values corresponding to inputs
    input_signals = _signals.get_signal_list(widget_class, "inputs")
    for signal in input_signals:
        if signal.ewoksname not in inputs:
            continue
        used_values.add(signal.ewoksname)
        value = value_from_transfer(inputs[signal.ewoksname])
        input_list.append((signal, value))

    # Node properties not corresponding to settings or inputs
    # are used in settings migration
    unused_values = set(inputs.keys()) - used_values
    for ewoksname in unused_values:
        value = value_from_transfer(inputs[ewoksname])
        settings_dict[ewoksname] = value

    return input_list, settings_dict



from ewoksorange.gui.orange_utils.signals import Input
from ewoksorange.gui.orange_utils.signals import Output
from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ewoksorange.gui.widgets.simple_types_mixin import IntegerAdderMixin
from ewoksorange.tests.examples.tasks import SumTaskTest


class OWSumTaskTest(
    IntegerAdderMixin, OWEwoksWidgetNoThread, ewokstaskclass=SumTaskTest
):
    name = "SumTaskTest"
    description = "Adds two numbers"
    icon = "icons/sum.png"
    want_main_area = True

    class Inputs:
        a = Input("A", object)
        b = Input("B", object)

    class Outputs:
        result = Output("A + B", object)



from ewokscore.tests.examples.tasks.sumtask import SumTask

from ewoksorange.gui.orange_utils.signals import Input
from ewoksorange.gui.orange_utils.signals import Output
from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetOneThread
from ewoksorange.gui.widgets.simple_types_mixin import IntegerAdderMixin

__all__ = ["SumTask"]


class OWSumTask(IntegerAdderMixin, OWEwoksWidgetOneThread, ewokstaskclass=SumTask):
    name = "SumTask"
    description = "Adds two numbers"
    icon = "icons/sum.png"
    want_main_area = True

    class Inputs:
        a = Input("A", object)
        b = Input("B", object)
        c = Input("C", int)

    class Outputs:
        result = Output("A + B", object)



try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files

import pytest
from ewokscore import load_graph
from ewokscore.tests.examples.graphs import get_graph
from ewokscore.tests.examples.graphs import graph_names

from ..gui.workflows.owscheme import ewoks_to_ows
from ..gui.workflows.owscheme import graph_is_supported
from ..gui.workflows.owscheme import ows_to_ewoks


def test_ows_to_ewoks_sumtask_tutorial(tmpdir):
    """Test conversion of orange worflow files to ewoks and back"""
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumtask_tutorial.ows")
    ewoksgraph = ows_to_ewoks(str(filename))

    destination = str(tmpdir / "ewoksgraph.ows")
    ewoks_to_ows(ewoksgraph, destination, error_on_duplicates=False)
    ewoksgraph2 = ows_to_ewoks(destination)
    assert ewoksgraph == ewoksgraph2


def test_ows_to_ewoks_sumlist_tutorial(tmpdir):
    """Test conversion of orange worflow files to ewoks and back"""
    from orangecontrib.ewokstest import tutorials

    filename = resource_files(tutorials).joinpath("sumlist_tutorial.ows")
    ewoksgraph = ows_to_ewoks(str(filename))

    destination = str(tmpdir / "ewoksgraph.ows")
    ewoks_to_ows(ewoksgraph, destination)
    ewoksgraph2 = ows_to_ewoks(destination)
    assert ewoksgraph == ewoksgraph2


@pytest.mark.parametrize("graph_name", graph_names())
def test_ewoks_to_ows(graph_name, tmpdir):
    """Test conversion of ewoks to orange worflow files and back"""
    graph, _ = get_graph(graph_name)
    ewoksgraph = load_graph(graph)
    ewoksgraph.graph.graph.pop("ows", None)
    for node_id, node_attrs in ewoksgraph.graph.nodes.items():
        node_attrs["label"] = node_id
        node_attrs.pop("ows", None)
        node_attrs.pop("uiProps", None)

    destination = str(tmpdir / "ewoksgraph2.ows")
    if not graph_is_supported(ewoksgraph):
        with pytest.raises(RuntimeError):
            ewoks_to_ows(ewoksgraph, destination)
        return

    ewoks_to_ows(ewoksgraph, destination, error_on_duplicates=False)
    ewoksgraph2 = ows_to_ewoks(
        destination, title_as_node_id=True, preserve_ows_info=False
    )
    assert ewoksgraph == ewoksgraph2



from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread
from ewoksorange.tests.examples.tasks import PrintSum
from ewoksorange.tests.examples.widgets import WidgetMixin


class PrintSumOW(WidgetMixin, OWEwoksWidgetNoThread, ewokstaskclass=PrintSum):
    name = "Print list sum"
    description = "Print received list sum"
    icon = "icons/mywidget.svg"
    want_main_area = True



from __future__ import annotations

from typing import Any

from pydantic import BaseModel
from pydantic_core import PydanticUndefined


def _get_model_default_values(model: type[BaseModel]) -> dict[str, Any]:
    """
    Docstring for _get_model_default_values

    :return: Dict of input name -> value or missing marker.
    """
    field_with_values = filter(
        lambda pair: pair[1] is not PydanticUndefined,
        _get_default_field_factory(model=model),
    )
    return dict(field_with_values)


def _get_default_field_factory(model):
    for field_name, field in model.model_fields.items():
        yield field_name, field.default



from ewokscore import missing_data

from ...gui.widgets.simple_types_mixin import SimpleTypesWidgetMixin


def serialize_list(value) -> str:
    if missing_data.is_missing_data(value):
        return value
    else:
        return f"<List length={len(value)}>"


class WidgetMixin(SimpleTypesWidgetMixin):
    def _get_parameter_options(self, name):
        if name == "list":
            return {"serialize": serialize_list}
        elif name == "sum":
            return {"value_for_type": 0.0}
        elif name == "length":
            return {"value_for_type": 0}
        else:
            return dict()



from queue import Queue
from typing import Iterable

from AnyQt.QtCore import QObject
from AnyQt.QtCore import pyqtSignal as Signal

from ..qt_utils.signals import block_signals
from .threaded import ThreadedTaskExecutor


class TaskExecutorQueue(QObject, Queue):
    """Processing Queue with a First In, First Out behavior"""

    sigComputationStarted = Signal()
    """Signal emitted when a computation is started"""
    sigComputationEnded = Signal()
    """Signal emitted when a computation is ended"""

    def __init__(self, ewokstaskclass):
        super().__init__()
        self._task_executor = _ThreadedTaskExecutor(ewokstaskclass=ewokstaskclass)
        self._task_executor.finished.connect(self._process_ended)
        self._available = True
        """Simple thread to know if we can do some processing
        and avoid to mix thinks with QSignals and different threads
        """

    @property
    def is_available(self) -> bool:
        return self._available

    def add(self, **kwargs):
        """Add a task `ewokstaskclass` execution request"""
        super().put(kwargs)
        if self.is_available:
            self._process_next()

    def _process_next(self):
        if Queue.empty(self):
            return
        self._available = False
        self._task_executor.create_task(**Queue.get(self))
        if self._task_executor.has_task:
            self.sigComputationStarted.emit()
            self._task_executor.start()
        else:
            self._task_executor.finished.emit()

    def _process_ended(self):
        self._process_ended_direct(self.sender())

    def _process_ended_direct(self, task_executor: "_ThreadedTaskExecutor"):
        for callback in task_executor.callbacks:
            callback()
        self.sigComputationEnded.emit()
        self._available = True
        if self.is_available:
            self._process_next()

    def cancel_running_task(self, wait=True):
        """
        will cancel current task.
        task_executor signal 'finished' will be blocked but callbacks will be executed to ensure a safe processing
        """
        if not self.is_available:
            with block_signals(self._task_executor):
                self._task_executor.cancel_running_task()
                # stop and remove the current task from the stack
                self._task_executor.stop(wait=wait)
                # signal that processing is done
                self._process_ended_direct(task_executor=self._task_executor)

    def stop(self):
        """
        stop the queue. Wait for the last processing to be finished and reset current_task
        """
        self._task_executor.finished.disconnect(self._process_ended)
        while not self.empty():
            self.get()
        self._task_executor.stop(wait=True)
        self._task_executor = None

    @property
    def current_task(self):
        return self._task_executor.current_task


class _ThreadedTaskExecutor(ThreadedTaskExecutor):
    """Processing thread with some information on callbacks to be executed"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__callbacks = tuple()

    def create_task(
        self,
        _callbacks: Iterable = tuple(),
        _log_missing_inputs: bool = False,
        **kwargs,
    ):
        kwargs["log_missing_inputs"] = _log_missing_inputs
        super().create_task(**kwargs)
        self.__callbacks = _callbacks

    @property
    def callbacks(self):
        """Methods to be executed by the thread once the computation is done"""
        return self.__callbacks



import logging
import os
from contextlib import contextmanager
from typing import Iterator
from typing import Optional
from typing import Sequence
from typing import Tuple

import h5py
import silx.io
from silx.gui import icons
from silx.gui import qt
from silx.gui.data.DataViewerFrame import DataViewerFrame
from silx.gui.hdf5 import Hdf5ContextMenuEvent
from silx.gui.hdf5 import Hdf5TreeModel
from silx.gui.hdf5 import Hdf5TreeView
from silx.gui.hdf5 import NexusSortFilterProxyModel

_logger = logging.getLogger(__name__)


class DataViewer(qt.QWidget):
    """Browse data from files supported by silx.

    To create the widget

    .. code: python

        viewer = DataViewer(parent)
        viewer.setVisible(True)
        parent.layout().addWidget(viewer)

    To close and refresh files

    .. code: python

        viewer.updateFile("/path/to/file1.h5")
        viewer.updateFile("/path/to/file2.h5")
        viewer.closeFile("/path/to/file1.h5")

    To close all files

    .. code: python

        viewer.closeAll()
    """

    def __init__(self, parent):
        super().__init__(parent)

        self._h5files = list()

        # Do we need buttons for these?
        # silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION = "downward"
        # silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION = "upward"
        # silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"
        # silx.config.DEFAULT_PLOT_BACKEND = "opengl"

        self.__treePanel = qt.QSplitter(self)
        self.__treePanel.setOrientation(qt.Qt.Vertical)

        self.__dataPanel = DataViewerFrame(self)

        self.__treeview = Hdf5TreeView(self)
        self.__treeview.setExpandsOnDoubleClick(False)

        self.__treeWindow = self.__createTreeWindow(self.__treeview)
        self.__treeModelSorted = self.__createTreeModel(self.__treeview)

        self.__treePanel.addWidget(self.__treeWindow)
        self.__treePanel.setStretchFactor(1, 1)
        self.__treePanel.setCollapsible(0, False)

        self.__mainWidget = self.__createMainWidget(self.__treePanel, self.__dataPanel)

        self.__setLayout(self.__mainWidget)

        self.__treeview.activated.connect(self.displaySelectedData)
        self.__treeview.addContextMenuCallback(self.treeContextMenu)
        self.__customizeTreeModelColumns()

    def __setLayout(self, mainWidget):
        layout = qt.QVBoxLayout()
        layout.addWidget(mainWidget)
        layout.setStretchFactor(mainWidget, 1)
        self.setLayout(layout)

    def __createMainWidget(self, *widgets) -> qt.QWidget:
        mainWidget = qt.QSplitter(self)
        for widget in widgets:
            mainWidget.addWidget(widget)
        mainWidget.setStretchFactor(1, 1)
        for i in range(len(widgets)):
            mainWidget.setCollapsible(i, False)
        return mainWidget

    def __createTreeModel(self, treeview: Hdf5TreeView) -> NexusSortFilterProxyModel:
        treeModel = Hdf5TreeModel(treeview, ownFiles=False)
        treeModel.sigH5pyObjectLoaded.connect(self.__h5FileLoaded)
        treeModel.sigH5pyObjectRemoved.connect(self.__h5FileRemoved)
        treeModel.sigH5pyObjectSynchronized.connect(self.__h5FileSynchronized)
        treeModel.setDatasetDragEnabled(True)

        treeModelSorted = NexusSortFilterProxyModel(treeview)
        treeModelSorted.setSourceModel(treeModel)
        treeModelSorted.sort(0, qt.Qt.AscendingOrder)
        treeModelSorted.setSortCaseSensitivity(qt.Qt.CaseInsensitive)
        treeview.setModel(treeModelSorted)
        return treeModelSorted

    def __customizeTreeModelColumns(self):
        treeModel = self.__treeview.findHdf5TreeModel()
        columns = list(treeModel.COLUMN_IDS)
        columns.remove(treeModel.VALUE_COLUMN)
        columns.remove(treeModel.NODE_COLUMN)
        columns.remove(treeModel.DESCRIPTION_COLUMN)
        columns.insert(1, treeModel.DESCRIPTION_COLUMN)
        self.__treeview.header().setSections(columns)

    def __createTreeWindow(self, treeView: Hdf5TreeView) -> qt.QWidget:
        toolbar = qt.QToolBar(self)
        toolbar.setIconSize(qt.QSize(16, 16))
        toolbar.setStyleSheet("QToolBar { border: 0px }")

        action = qt.QAction(toolbar)
        action.setIcon(icons.getQIcon("view-refresh"))
        action.setText("Refresh")
        action.setToolTip("Refresh all selected items")
        action.triggered.connect(self.__refreshSelected)
        action.setShortcut(qt.QKeySequence(qt.Qt.Key_F5))
        toolbar.addAction(action)
        treeView.addAction(action)
        self.__refreshAction = action

        # Another shortcut for refresh
        action = qt.QAction(toolbar)
        action.setShortcut(qt.QKeySequence(qt.Qt.CTRL | qt.Qt.Key_R))
        treeView.addAction(action)
        action.triggered.connect(self.__refreshSelected)

        action = qt.QAction(toolbar)
        # action.setIcon(icons.getQIcon("view-refresh"))
        action.setText("Close")
        action.setToolTip("Close selected item")
        action.triggered.connect(self.__removeSelected)
        action.setShortcut(qt.QKeySequence(qt.Qt.Key_Delete))
        treeView.addAction(action)
        self.__closeAction = action

        toolbar.addSeparator()

        action = qt.QAction(toolbar)
        action.setIcon(icons.getQIcon("tree-expand-all"))
        action.setText("Expand all")
        action.setToolTip("Expand all selected items")
        action.triggered.connect(self.__expandAllSelected)
        action.setShortcut(qt.QKeySequence(qt.Qt.CTRL | qt.Qt.Key_Plus))
        toolbar.addAction(action)
        treeView.addAction(action)
        self.__expandAllAction = action

        action = qt.QAction(toolbar)
        action.setIcon(icons.getQIcon("tree-collapse-all"))
        action.setText("Collapse all")
        action.setToolTip("Collapse all selected items")
        action.triggered.connect(self.__collapseAllSelected)
        action.setShortcut(qt.QKeySequence(qt.Qt.CTRL | qt.Qt.Key_Minus))
        toolbar.addAction(action)
        treeView.addAction(action)
        self.__collapseAllAction = action

        action = qt.QAction("&Sort file content", toolbar)
        action.setIcon(icons.getQIcon("tree-sort"))
        action.setToolTip("Toggle sorting of file content")
        action.setCheckable(True)
        action.setChecked(True)
        action.triggered.connect(self.setContentSorted)
        toolbar.addAction(action)
        treeView.addAction(action)
        self._sortContentAction = action

        widget = qt.QWidget(self)
        layout = qt.QVBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(toolbar)
        layout.addWidget(treeView)
        return widget

    def __iterModelIndices(
        self, max_depth: Optional[int] = None, indexes: Optional[Sequence] = None
    ) -> Iterator[Tuple[Tuple[qt.QModelIndex, int], qt.QModelIndex, int]]:
        selection = self.__treeview.selectionModel()
        if indexes is None:
            indexes = selection.selectedIndexes()
        while len(indexes) > 0:
            index = indexes.pop(0)
            if isinstance(index, tuple):
                index, depth = index
            else:
                depth = 0
            if index.column() != 0:
                continue
            if max_depth is not None and depth > max_depth:
                break
            yield indexes, index, depth

    @staticmethod
    def __getRootIndex(index: qt.QModelIndex):
        rootIndex = index
        while rootIndex.parent().isValid():
            rootIndex = rootIndex.parent()
        return rootIndex

    def __removeSelected(self):
        """Close selected items"""
        model = self.__treeview.model()
        h5files = set()
        selectedItems = []
        with self.__waitCursor():
            for _, index, _ in self.__iterModelIndices():
                rootIndex = self.__getRootIndex(index)
                relativePath = self.__getRelativePath(model, rootIndex, index)
                selectedItems.append((rootIndex.row(), relativePath))

                h5 = model.data(index, role=Hdf5TreeModel.H5PY_OBJECT_ROLE)
                h5files.add(h5.file)

            if not h5files:
                return

            model = self.__treeview.findHdf5TreeModel()
            for h5 in h5files:
                model.removeH5pyObject(h5)

    def __refreshSelected(self):
        """Refresh all selected items"""
        model = self.__treeview.model()
        selection = self.__treeview.selectionModel()
        selectedItems = []
        h5files = []
        with self.__waitCursor():
            for _, index, _ in self.__iterModelIndices():
                rootIndex = self.__getRootIndex(index)
                relativePath = self.__getRelativePath(model, rootIndex, index)
                selectedItems.append((rootIndex.row(), relativePath))

                h5 = model.data(rootIndex, role=Hdf5TreeModel.H5PY_OBJECT_ROLE)
                item = model.data(rootIndex, role=Hdf5TreeModel.H5PY_ITEM_ROLE)
                h5files.append((h5, item._openedPath))

            if not h5files:
                return

            for h5, filename in h5files:
                self.__synchronizeH5pyObject(h5, filename)

            itemSelection = qt.QItemSelection()
            for rootRow, relativePath in selectedItems:
                rootIndex = model.index(rootRow, 0, qt.QModelIndex())
                index = self.__indexFromPath(model, rootIndex, relativePath)
                if index is None:
                    continue
                indexEnd = model.index(
                    index.row(), model.columnCount() - 1, index.parent()
                )
                itemSelection.select(index, indexEnd)
            selection.select(itemSelection, qt.QItemSelectionModel.ClearAndSelect)

    def __synchronizeH5pyObject(self, h5, filename: Optional[str] = None):
        model = self.__treeview.findHdf5TreeModel()
        # This is buggy right now while h5py do not allow to close a file
        # while references are still used.
        # FIXME: The architecture have to be reworked to support this feature.
        # model.synchronizeH5pyObject(h5)

        if filename is None:
            filename = f"{h5.file.filename}::{h5.name}"
        row = model.h5pyObjectRow(h5)
        index = self.__treeview.model().index(row, 0, qt.QModelIndex())
        paths = self.__getPathFromExpandedNodes(self.__treeview, index)
        model.removeH5pyObject(h5)
        model.insertFile(filename, row)
        index = self.__treeview.model().index(row, 0, qt.QModelIndex())
        self.__expandNodesFromPaths(self.__treeview, index, paths)

    def __getRelativePath(self, model, rootIndex, index):
        """Returns a relative path from an index to his rootIndex.

        If the path is empty the index is also the rootIndex.
        """
        path = ""
        while index.isValid():
            if index == rootIndex:
                return path
            name = model.data(index)
            if path == "":
                path = name
            else:
                path = name + "/" + path
            index = index.parent()

        # index is not a children of rootIndex
        raise ValueError("index is not a children of the rootIndex")

    def __getPathFromExpandedNodes(self, view, rootIndex):
        """Return relative path from the root index of the extended nodes"""
        model = view.model()
        rootPath = None
        paths = []

        for indexes, index, depth in self.__iterModelIndices(indexes=[rootIndex]):
            if not view.isExpanded(index):
                continue

            node = model.data(index, role=Hdf5TreeModel.H5PY_ITEM_ROLE)
            path = node._getCanonicalName()
            if rootPath is None:
                rootPath = path
            path = path[len(rootPath) :]
            paths.append(path)

            for child in range(model.rowCount(index)):
                childIndex = model.index(child, 0, index)
                indexes.append((childIndex, depth + 1))
        return paths

    def __indexFromPath(self, model, rootIndex, path):
        elements = path.split("/")
        if elements[0] == "":
            elements.pop(0)
        index = rootIndex
        while len(elements) != 0:
            element = elements.pop(0)
            found = False
            for child in range(model.rowCount(index)):
                childIndex = model.index(child, 0, index)
                name = model.data(childIndex)
                if element == name:
                    index = childIndex
                    found = True
                    break
            if not found:
                return None
        return index

    def __expandNodesFromPaths(self, view, rootIndex, paths):
        model = view.model()
        for path in paths:
            index = self.__indexFromPath(model, rootIndex, path)
            if index is not None:
                view.setExpanded(index, True)

    @contextmanager
    def __waitCursor(self):
        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        try:
            yield
        finally:
            qt.QApplication.restoreOverrideCursor()

    def __expandAllSelected(self):
        """Expand all selected items of the tree."""
        with self.__waitCursor():
            self.__setExpanded(True)

    def __collapseAllSelected(self):
        """Collapse all selected items of the tree."""
        self.__setExpanded(False)

    def __setExpanded(self, expanded: bool):
        model = self.__treeview.model()
        for indexes, index, depth in self.__iterModelIndices(max_depth=2):
            if not model.hasChildren(index):
                continue
            self.__treeview.setExpanded(index, expanded)
            for row in range(model.rowCount(index)):
                childIndex = model.index(row, 0, index)
                indexes.append((childIndex, depth + 1))

    def __h5FileLoaded(self, loadedH5):
        self._h5files.append(loadedH5)
        self.displayData(loadedH5)

    def __h5FileRemoved(self, removedH5):
        data = self.__dataPanel.data()
        if data is not None:
            if data.file is not None:
                if data.file.filename == removedH5.file.filename:
                    self.__dataPanel.setData(None)
        removedH5.close()
        self._h5files.remove(removedH5)

    def __h5FileSynchronized(self, removedH5, loadedH5):
        data = self.__dataPanel.data()
        if data is not None:
            if data.file is not None:
                if data.file.filename == removedH5.file.filename:
                    try:
                        newData = loadedH5[data.name]
                        self.__dataPanel.setData(newData)
                    except Exception:
                        _logger.debug("Cannot synchronize", exc_info=True)
        removedH5.close()
        self._h5files.remove(removedH5)

    def closeEvent(self, event):
        self.displayData(None)
        self.closeAll()

    def closeAll(self):
        """Close all currently opened files"""
        self.__treeview.findHdf5TreeModel().clear()

    def __getFileObject(self, filename):
        for h5file in self._h5files:
            if h5file.filename == filename:
                return h5file

    def closeFile(self, filename):
        h5file = self.__getFileObject(filename)
        model = self.__treeview.findHdf5TreeModel()
        model.removeH5pyObject(h5file)

    def updateFile(self, filename):
        if not os.path.exists(filename):
            return
        h5file = self.__getFileObject(filename)
        if h5file is not None:
            self.__refreshAction.trigger()
            return
        self.closeFile(filename)
        model = self.__treeview.findHdf5TreeModel()
        h5file = h5py.File(filename, mode="a")
        try:
            model.sigH5pyObjectLoaded.emit(h5file, filename)
        except TypeError:
            # Support silx<2.0.0
            model.sigH5pyObjectLoaded.emit(h5file)
        model.insertH5pyObject(h5file, filename=filename)

    def setContentSorted(self, sort):
        """Set whether file content should be sorted or not.

        :param bool sort:
        """
        sort = bool(sort)
        if sort != self.isContentSorted():
            # save expanded nodes
            pathss = []
            root = qt.QModelIndex()
            model = self.__treeview.model()
            for i in range(model.rowCount(root)):
                index = model.index(i, 0, root)
                paths = self.__getPathFromExpandedNodes(self.__treeview, index)
                pathss.append(paths)

            self.__treeview.setModel(
                self.__treeModelSorted if sort else self.__treeModelSorted.sourceModel()
            )
            self._sortContentAction.setChecked(self.isContentSorted())

            # restore expanded nodes
            model = self.__treeview.model()
            for i in range(model.rowCount(root)):
                index = model.index(i, 0, root)
                paths = pathss.pop(0)
                self.__expandNodesFromPaths(self.__treeview, index, paths)

    def isContentSorted(self):
        """Returns whether the file content is sorted or not.

        :rtype: bool
        """
        return self.__treeview.model() is self.__treeModelSorted

    def displaySelectedData(self):
        """Called to update the dataviewer with the selected data."""
        selected = list(self.__treeview.selectedH5Nodes(ignoreBrokenLinks=False))
        if len(selected) == 1:
            # Update the viewer for one selection
            data = selected[0]
            self.__dataPanel.setData(data)
        else:
            _logger.debug("Too many data selected")

    def displayData(self, data):
        """Called to update the dataviewer with specific data."""
        self.__dataPanel.setData(data)

    def treeContextMenu(self, event: Hdf5ContextMenuEvent):
        """Called to populate the context menu"""
        selectedObjects = event.source().selectedH5Nodes(ignoreBrokenLinks=False)
        menu = event.menu()

        if not menu.isEmpty():
            menu.addSeparator()

        for obj in selectedObjects:
            h5 = obj.h5py_object

            name = obj.name
            if name.startswith("/"):
                name = name[1:]
            if name == "":
                name = "the root"

            action = qt.QAction("Show %s" % name, event.source())
            action.triggered.connect(lambda: self.displayData(h5))
            menu.addAction(action)

            if silx.io.is_file(h5):
                action = qt.QAction("Close %s" % obj.local_filename, event.source())
                action.triggered.connect(
                    lambda: self.__treeview.findHdf5TreeModel().removeH5pyObject(h5)
                )
                menu.addAction(action)
                action = qt.QAction(
                    "Synchronize %s" % obj.local_filename, event.source()
                )
                action.triggered.connect(lambda: self.__synchronizeH5pyObject(h5))
                menu.addAction(action)
